# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Schildkroet                                        *
# *   Copyright (c) 2023 luvtofish                                          *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENSE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


import FreeCAD
import Part
import Path
import Path.Base.FeedRate as PathFeedRate
import Path.Base.Generator.tapping as tapping
import Path.Base.MachineState as PathMachineState
import Path.Op.Base as PathOp
import Path.Op.CircularHoleBase as PathCircularHoleBase
import PathScripts.PathUtils as PathUtils
from PySide.QtCore import QT_TRANSLATE_NOOP

__title__ = "Path Tapping Operation"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Path Tapping operation."
__contributors__ = "luvtofish (Dan Henderson)"

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class ObjectTapping(PathCircularHoleBase.ObjectOp):
    """Proxy object for Tapping  operation."""

    @classmethod
    def propertyEnumerations(self, dataType="data"):
        """helixOpPropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        # Enumeration lists for App::PropertyEnumeration properties
        enums = {
            "ReturnLevel": [
                (translate("CAM_Tapping", "G98"), "G98"),
                (translate("CAM_Tapping", "G99"), "G99"),
            ],  # How high to retract after a tapping move
            "ExtraOffset": [
                (translate("CAM_Tapping", "None"), "None"),
                (translate("CAM_Tapping", "Drill Tip"), "Drill Tip"),
                (translate("CAM_Tapping", "2x Drill Tip"), "2x Drill Tip"),
            ],  # extra drilling depth to clear tap taper
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def circularHoleFeatures(self, obj):
        """circularHoleFeatures(obj) ... tapping works on anything, turn on all Base geometries and Locations."""
        return PathOp.FeatureBaseGeometry | PathOp.FeatureLocations | PathOp.FeatureCoolant

    def initCircularHoleOperation(self, obj):
        """initCircularHoleOperation(obj) ... add tapping specific properties to obj."""
        obj.addProperty(
            "App::PropertyFloat",
            "DwellTime",
            "Tap",
            QT_TRANSLATE_NOOP("App::Property", "The time to dwell at bottom of tapping cycle"),
        )
        obj.addProperty(
            "App::PropertyBool",
            "DwellEnabled",
            "Tap",
            QT_TRANSLATE_NOOP("App::Property", "Enable dwell"),
        )
        obj.addProperty(
            "App::PropertyBool",
            "AddTipLength",
            "Tap",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Calculate the tip length and subtract from final depth",
            ),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "ReturnLevel",
            "Tap",
            QT_TRANSLATE_NOOP("App::Property", "Controls how tool retracts Default=G98"),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "RetractHeight",
            "Tap",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The height where feed starts and height during retract tool when path is finished while in a peck operation",
            ),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "ExtraOffset",
            "Tap",
            QT_TRANSLATE_NOOP("App::Property", "How far the tap depth is extended"),
        )

        for n in self.propertyEnumerations():
            setattr(obj, n[0], n[1])

    def circularHoleExecute(self, obj, holes):
        """circularHoleExecute(obj, holes) ... generate tapping operation for each hole in holes."""
        Path.Log.track()
        machine = PathMachineState.MachineState()

        if not hasattr(obj.ToolController.Tool, "Pitch"):
            Path.Log.error(
                translate(
                    "Path_Tapping",
                    "Tapping Operation requires a Tap tool with Pitch",
                )
            )
            return

        self.commandlist.append(Path.Command("(Begin Tapping)"))

        # rapid to clearance height
        command = Path.Command("G0", {"Z": obj.ClearanceHeight.Value})
        machine.addCommand(command)
        self.commandlist.append(command)

        self.commandlist.append(Path.Command("G90"))  # Absolute distance mode

        # Calculate offsets to add to target edge
        endoffset = 0.0
        if obj.ExtraOffset == "Drill Tip":
            endoffset = PathUtils.drillTipLength(self.tool)
        elif obj.ExtraOffset == "2x Drill Tip":
            endoffset = PathUtils.drillTipLength(self.tool) * 2

        # http://linuxcnc.org/docs/html/gcode/g-code.html#gcode:g98-g99
        self.commandlist.append(Path.Command(obj.ReturnLevel))

        # This section is technical debt. The computation of the
        # target shapes should be factored out for reuse.
        # This will likely mean refactoring upstream CircularHoleBase to pass
        # spot shapes instead of holes.

        startHeight = obj.StartDepth.Value + self.job.SetupSheet.SafeHeightOffset.Value

        edgelist = []
        for hole in holes:
            v1 = FreeCAD.Vector(hole["x"], hole["y"], obj.StartDepth.Value)
            v2 = FreeCAD.Vector(hole["x"], hole["y"], obj.FinalDepth.Value - endoffset)
            edgelist.append(Part.makeLine(v1, v2))

        # iterate the edgelist and generate gcode
        for edge in edgelist:
            Path.Log.debug(edge)

            # move to hole location

            startPoint = edge.Vertexes[0].Point

            command = Path.Command(
                "G0", {"X": startPoint.x, "Y": startPoint.y, "Z": startHeight}
            )  # DLH added Z offset to fix issue with Dynapath Control.
            self.commandlist.append(command)
            machine.addCommand(command)

            # command = Path.Command("G0", {"Z": startHeight})
            # self.commandlist.append(command)
            # machine.addCommand(command)

            # command = Path.Command("G1", {"Z": obj.StartDepth.Value})
            # self.commandlist.append(command)
            # machine.addCommand(command)

            # Technical Debt:  We are assuming the edges are aligned.
            # This assumption should be corrected and the necessary rotations
            # performed to align the edge with the Z axis for drilling

            # Perform tapping
            dwelltime = obj.DwellTime if obj.DwellEnabled else 0.0
            repeat = 1  # technical debt:  Add a repeat property for user control

            # Get attribute from obj.tool, assign default and set to bool for passing to generate
            isRightHand = (
                getattr(obj.ToolController.Tool, "SpindleDirection", "Forward") == "Forward"
            )

            # Get pitch in mm as a float (no unit string)
            pitch = getattr(obj.ToolController.Tool, "Pitch", None)
            if pitch is None or pitch == 0:
                Path.Log.error(
                    translate(
                        "Path_Tapping",
                        "Tapping Operation requires a Tap tool with non-zero Pitch",
                    )
                )
                continue

            spindle_speed = getattr(obj.ToolController, "SpindleSpeed", None)
            if spindle_speed is None or spindle_speed == 0:
                Path.Log.error(
                    translate(
                        "Path_Tapping",
                        "Tapping Operation requires a ToolController with non-zero SpindleSpeed",
                    )
                )
                continue

            try:
                tappingcommands = tapping.generate(
                    edge,
                    dwelltime,
                    repeat,
                    obj.RetractHeight.Value,
                    isRightHand,
                    pitch,
                    spindle_speed,
                )

            except ValueError as e:  # any targets that fail the generator are ignored
                Path.Log.info(e)
                continue

            for command in tappingcommands:
                self.commandlist.append(command)
                machine.addCommand(command)

        # Cancel canned tapping cycle
        self.commandlist.append(Path.Command("G80"))
        # command = Path.Command("G0", {"Z": obj.SafeHeight.Value})  DLH- Not needed, adds unnecessary move to Z SafeHeight.
        # self.commandlist.append(command)
        # machine.addCommand(command)       DLH - Not needed.

        # Apply feed rates to commands
        PathFeedRate.setFeedRate(self.commandlist, obj.ToolController)

    def opSetDefaultValues(self, obj, job):
        """opSetDefaultValues(obj, job) ... set default value for RetractHeight"""
        obj.ExtraOffset = "None"

        if hasattr(job.SetupSheet, "RetractHeight"):
            obj.RetractHeight = job.SetupSheet.RetractHeight
        elif self.applyExpression(obj, "RetractHeight", "StartDepth+SetupSheet.SafeHeightOffset"):
            if not job:
                obj.RetractHeight = 10
            else:
                obj.RetractHeight.Value = obj.StartDepth.Value + 1.0

        if hasattr(job.SetupSheet, "DwellTime"):
            obj.DwellTime = job.SetupSheet.DwellTime
        else:
            obj.DwellTime = 1


def SetupProperties():
    setup = []
    setup.append("DwellTime")
    setup.append("DwellEnabled")
    setup.append("AddTipLength")
    setup.append("ReturnLevel")
    setup.append("ExtraOffset")
    setup.append("RetractHeight")
    return setup


def Create(name, obj=None, parentJob=None):
    """Create(name) ... Creates and returns a Tapping operation."""
    if obj is None:
        obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)

    obj.Proxy = ObjectTapping(obj, name, parentJob)
    if obj.Proxy:
        obj.Proxy.findAllHoles(obj)

    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later

from .camotics import CamoticsLibrarySerializer
from .fctl import FCTLSerializer
from .linuxcnc import LinuxCNCSerializer


all_serializers = CamoticsLibrarySerializer, FCTLSerializer, LinuxCNCSerializer


__all__ = [
    "CamoticsLibrarySerializer",
    "FCTLSerializer",
    "LinuxCNCSerializer",
]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2023 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Unit tests for the Draft Workbench, array tests."""

## @package test_array
# \ingroup drafttests
# \brief Unit tests for the Draft Workbench, array tests.

## \addtogroup drafttests
# @{

import Draft
from FreeCAD import Vector
from drafttests import test_base


class DraftArray(test_base.DraftTestCaseDoc):
    """Test Draft array functions."""

    def test_link_array(self):
        """Create a link array."""
        box = self.doc.addObject("Part::Box", "Box")
        box.Label = "Box"
        self.doc.recompute()

        array = Draft.make_ortho_array(
            box,
            v_x=Vector(100.0, 0.0, 0.0),
            v_y=Vector(0.0, 100.0, 0.0),
            v_z=Vector(0.0, 0.0, 100.0),
            n_x=12,
            n_y=1,
            n_z=1,
            use_link=True,
        )

        Draft.autogroup(array)
        array.ExpandArray = True
        array.Fuse = False
        self.doc.recompute(None, True, True)

        array.NumberX = 6
        self.doc.recompute(None, True, True)
        self.assertEqual(array.Count, array.NumberX)

        array.NumberX = 24
        self.doc.recompute(None, True, True)
        self.assertEqual(array.Count, array.NumberX)

        self.doc.recompute(None, True, True)
        self.assertEqual(array.Count, array.NumberX)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *   Copyright (c) 2024 The FreeCAD Project Association                    *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to rotate shapes around a center and axis."""
## @package rotate
# \ingroup draftfunctions
# \brief Provides functions to rotate shapes around a center and axis.

## \addtogroup draftfunctions
# @{
import math

import FreeCAD as App
import DraftVecUtils
from draftfunctions import join
from draftmake import make_copy
from draftmake import make_line
from draftobjects import layer
from draftutils import groups
from draftutils import gui_utils
from draftutils import params
from draftutils import utils


def rotate(
    selection,
    angle,
    center=App.Vector(0, 0, 0),
    axis=App.Vector(0, 0, 1),
    copy=False,
    subelements=False,
):
    """rotate(selection, angle, [center], [axis], [copy], [subelements])

    Rotates or copies selected objects.

    Parameters
    ----------
    selection: single object / list of objects / selection set
        When dealing with nested objects, use `Gui.Selection.getSelectionEx("", 0)`
        to create the selection set.

    angle: float / integer
        Rotation angle (in degrees).

    center: App.Vector, optional
        Defaults to `App.Vector(0, 0, 0)`.
        Rotation center.

    axis: App.Vector, optional
        Defaults to `App.Vector(0, 0, 1)`.
        Rotation axis.

    copy: bool, optional
        Defaults to `False`.
        If `True` the selected objects are not rotated, but rotated copies are
        created instead.

    subelements: bool, optional
        Defaults to `False`.
        If `True` subelements instead of whole objects are processed.
        Only used if selection is a selection set.

    Returns
    -------
    single object / list with 2 or more objects / empty list
        The objects (or their copies).
    """
    utils.type_check(
        [
            (angle, (float, int)),
            (center, App.Vector),
            (axis, App.Vector),
            (copy, bool),
            (subelements, bool),
        ],
        "rotate",
    )
    if not isinstance(selection, list):
        selection = [selection]
    if not selection:
        return None

    if selection[0].isDerivedFrom("Gui::SelectionObject"):
        if subelements:
            return _rotate_subelements(selection, angle, center, axis, copy)
        else:
            objs, parent_places, sel_info = utils._modifiers_process_selection(selection, copy)
    else:
        objs = utils._modifiers_filter_objects(utils._modifiers_get_group_contents(selection), copy)
        parent_places = None
        sel_info = None

    if not objs:
        return None

    newobjs = []
    newgroups = {}

    if copy:
        for obj in objs:
            if obj.isDerivedFrom("App::DocumentObjectGroup") and obj.Name not in newgroups:
                newgroups[obj.Name] = obj.Document.addObject(
                    obj.TypeId, utils.get_real_name(obj.Name)
                )

    for idx, obj in enumerate(objs):
        newobj = None

        if parent_places is not None:
            parent_place = parent_places[idx]
        elif hasattr(obj, "getGlobalPlacement"):
            parent_place = obj.getGlobalPlacement() * obj.Placement.inverse()
        else:
            parent_place = App.Placement()

        if copy or parent_place.isIdentity():
            real_center = center
            real_axis = axis
        else:
            real_center = parent_place.inverse().multVec(center)
            real_axis = parent_place.inverse().Rotation.multVec(axis)

        if obj.isDerivedFrom("App::DocumentObjectGroup"):
            if copy:
                newobj = newgroups[obj.Name]
            else:
                newobj = obj

        elif hasattr(obj, "Placement"):
            if copy:
                newobj = make_copy.make_copy(obj)
                if not parent_place.isIdentity():
                    newobj.Placement = parent_place * newobj.Placement
            else:
                newobj = obj
            newobj.Placement.rotate(real_center, real_axis, angle, comp=True)

        elif obj.isDerivedFrom("App::Annotation"):
            if copy:
                newobj = make_copy.make_copy(obj)
                if not parent_place.isIdentity():
                    newobj.Position = parent_place.multVec(newobj.Position)
            else:
                newobj = obj
            newobj.Position = rotate_vector_from_center(newobj.Position, angle, axis, center)

        elif utils.get_type(obj) in ("Dimension", "LinearDimension", "AngularDimension"):
            # "Dimension" was the type for linear dimensions <= v0.18.
            if copy:
                newobj = make_copy.make_copy(obj)
                if not parent_place.isIdentity():
                    newobj.Proxy.transform(newobj, parent_place)
            else:
                newobj = obj
            pla = App.Placement()
            pla.rotate(real_center, real_axis, angle, comp=True)
            newobj.Proxy.transform(newobj, pla)

        if newobj is not None:
            newobjs.append(newobj)
            if copy:
                lyr = layer.get_layer(obj)
                if lyr is not None:
                    lyr.Proxy.addObject(lyr, newobj)
                for parent in obj.InList:
                    if parent.isDerivedFrom("App::DocumentObjectGroup") and (parent in objs):
                        newgroups[parent.Name].addObject(newobj)

    if not copy or params.get_param("selectBaseObjects"):
        if sel_info is not None:
            gui_utils.select(sel_info)
        else:
            gui_utils.select(objs)
    else:
        gui_utils.select(newobjs)

    if len(newobjs) == 1:
        return newobjs[0]
    return newobjs


def _rotate_subelements(selection, angle, center, axis, copy):
    data_list, sel_info = utils._modifiers_process_subselection(selection, copy)
    newobjs = []
    if copy:
        for obj, vert_idx, edge_idx, global_place in data_list:
            if edge_idx >= 0:
                newobjs.append(copy_rotated_edge(obj, edge_idx, angle, center, axis, global_place))
        newobjs = join.join_wires(newobjs)
    else:
        for obj, vert_idx, edge_idx, global_place in data_list:
            if vert_idx >= 0:
                rotate_vertex(obj, vert_idx, angle, center, axis, global_place)
            elif edge_idx >= 0:
                rotate_edge(obj, edge_idx, angle, center, axis, global_place)

    if not copy or params.get_param("selectBaseObjects"):
        gui_utils.select(sel_info)
    else:
        gui_utils.select(newobjs)

    if len(newobjs) == 1:
        return newobjs[0]
    return newobjs


def rotate_vector_from_center(vector, angle, axis, center):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    rv = vector.sub(center)
    rv = DraftVecUtils.rotate(rv, math.radians(angle), axis)
    return center.add(rv)


def rotate_vertex(obj, vert_idx, angle, center, axis, global_place=None):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    if global_place is None:
        glp = obj.getGlobalPlacement()
    else:
        glp = global_place
    points = obj.Points
    points[vert_idx] = glp.inverse().multVec(
        rotate_vector_from_center(glp.multVec(points[vert_idx]), angle, axis, center)
    )
    obj.Points = points


def rotate_edge(obj, edge_idx, angle, center, axis, global_place=None):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    rotate_vertex(obj, edge_idx, angle, center, axis, global_place)
    if utils.is_closed_edge(edge_idx, obj):
        rotate_vertex(obj, 0, angle, center, axis, global_place)
    else:
        rotate_vertex(obj, edge_idx + 1, angle, center, axis, global_place)


def copy_rotated_edge(obj, edge_idx, angle, center, axis, global_place=None):
    """
    Needed for SubObjects modifiers.
    Implemented by Dion Moult during 0.19 dev cycle (works only with Draft Wire).
    """
    if global_place is None:
        glp = obj.getGlobalPlacement()
    else:
        glp = global_place
    vertex1 = rotate_vector_from_center(glp.multVec(obj.Points[edge_idx]), angle, axis, center)
    if utils.is_closed_edge(edge_idx, obj):
        vertex2 = rotate_vector_from_center(glp.multVec(obj.Points[0]), angle, axis, center)
    else:
        vertex2 = rotate_vector_from_center(
            glp.multVec(obj.Points[edge_idx + 1]), angle, axis, center
        )
    newobj = make_line.make_line(vertex1, vertex2)
    gui_utils.format_object(newobj, obj)
    return newobj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path.Op.Custom as PathCustom
import Path.Op.Gui.Base as PathOpGui

from PySide.QtCore import QT_TRANSLATE_NOOP


__title__ = "CAM Custom Operation UI"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Custom operation page controller and command implementation."


class TaskPanelOpPage(PathOpGui.TaskPanelPage):
    """Page controller class for the Custom operation."""

    def getForm(self):
        """getForm() ... returns UI"""
        return FreeCADGui.PySideUic.loadUi(":/panels/PageOpCustomEdit.ui")

    def getFields(self, obj):
        """getFields(obj) ... transfers values from UI to obj's properties"""
        self.updateToolController(obj, self.form.toolController)
        self.updateCoolant(obj, self.form.coolantController)

    def setFields(self, obj):
        """setFields(obj) ... transfers obj's property values to UI"""
        self.setupToolController(obj, self.form.toolController)
        self.form.txtGCode.setText("\n".join(obj.Gcode))
        self.setupCoolant(obj, self.form.coolantController)

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals for updating obj"""
        signals = []
        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.coolantController.currentIndexChanged)
        self.form.txtGCode.textChanged.connect(self.setGCode)
        return signals

    def setGCode(self):
        self.obj.Gcode = self.form.txtGCode.toPlainText().split("\n")


Command = PathOpGui.SetupOperation(
    "Custom",
    PathCustom.Create,
    TaskPanelOpPage,
    "CAM_Custom",
    QT_TRANSLATE_NOOP("CAM_Custom", "Custom"),
    QT_TRANSLATE_NOOP("CAM_Custom", "Create custom G-code snippet"),
    PathCustom.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathCustomGui... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Helper functions that are used by IFC importer and exporter."""
import math

import FreeCAD
import Arch
import ArchIFC

from draftutils import params
from draftutils.messages import _msg, _wrn

if FreeCAD.GuiUp:
    import FreeCADGui as Gui


PREDEFINED_RGB = {
    "black": (0, 0, 0),
    "red": (1.0, 0, 0),
    "green": (0, 1.0, 0),
    "blue": (0, 0, 1.0),
    "yellow": (1.0, 1.0, 0),
    "magenta": (1.0, 0, 1.0),
    "cyan": (0, 1.0, 1.0),
    "white": (1.0, 1.0, 1.0),
}


DEBUG_prod_repr = False
DEBUG_prod_colors = False


def dd2dms(dd):
    """Convert decimal degrees to degrees, minutes, seconds.

    Used in export.
    """
    sign = 1 if dd >= 0 else -1
    dd = abs(dd)
    minutes, seconds = divmod(dd * 3600, 60)
    degrees, minutes = divmod(minutes, 60)

    if dd < 0:
        degrees = -degrees

    return (int(degrees) * sign, int(minutes) * sign, int(seconds) * sign)


def dms2dd(degrees, minutes, seconds, milliseconds=0):
    """Convert degrees, minutes, seconds to decimal degrees.

    Used in import.
    """
    dd = float(degrees) + float(minutes) / 60 + float(seconds) / 3600
    return dd


def getPreferences():
    """Retrieve the IFC preferences available in import and export.

    MERGE_MODE_ARCH:
        0 = parametric BIM objects
        1 = non-parametric BIM objects
        2 = Part shapes
        3 = One compound per storey
    """
    preferences = {
        "DEBUG": params.get_param_arch("ifcDebug"),
        "PREFIX_NUMBERS": params.get_param_arch("ifcPrefixNumbers"),
        "SKIP": params.get_param_arch("ifcSkip").split(","),
        "SEPARATE_OPENINGS": params.get_param_arch("ifcSeparateOpenings"),
        "ROOT_ELEMENT": params.get_param_arch("ifcRootElement"),
        "GET_EXTRUSIONS": params.get_param_arch("ifcGetExtrusions"),
        "MERGE_MATERIALS": params.get_param_arch("ifcMergeMaterials"),
        "MERGE_MODE_ARCH": params.get_param_arch("ifcImportModeArch"),
        "MERGE_MODE_STRUCT": params.get_param_arch("ifcImportModeStruct"),
        "CREATE_CLONES": params.get_param_arch("ifcCreateClones"),
        "IMPORT_PROPERTIES": params.get_param_arch("ifcImportProperties"),
        "SPLIT_LAYERS": params.get_param_arch(
            "ifcSplitLayers"
        ),  # wall layer, not layer for visual props
        "FITVIEW_ONIMPORT": params.get_param_arch("ifcFitViewOnImport"),
        "ALLOW_INVALID": params.get_param_arch("ifcAllowInvalid"),
        "REPLACE_PROJECT": params.get_param_arch("ifcReplaceProject"),
        "MULTICORE": params.get_param_arch("ifcMulticore"),
        "IMPORT_LAYER": params.get_param_arch("ifcImportLayer"),
    }

    if preferences["MERGE_MODE_ARCH"] > 0:
        preferences["SEPARATE_OPENINGS"] = False
        preferences["GET_EXTRUSIONS"] = False
    if not preferences["SEPARATE_OPENINGS"]:
        preferences["SKIP"].append("IfcOpeningElement")

    return preferences


class ProjectImporter:
    """A helper class to create an Arch Project object."""

    def __init__(self, file, objects):
        self.file = file
        self.objects = objects

    def execute(self):
        self.project = self.file.by_type("IfcProject")[0]
        self.object = Arch.makeProject()
        self.objects[self.project.id()] = self.object
        self.setAttributes()
        self.setComplexAttributes()

    def setAttributes(self):
        for prop in self.object.PropertiesList:
            if hasattr(self.project, prop) and getattr(self.project, prop):
                setattr(self.object, prop, getattr(self.project, prop))

    def setComplexAttributes(self):
        try:
            mapConversion = self.project.RepresentationContexts[0].HasCoordinateOperation[0]

            data = self.extractTargetCRSData(mapConversion.TargetCRS)
            data.update(self.extractMapConversionData(mapConversion))
            # TODO: review and refactor this piece of code.
            # Calling a method from a class is a bit strange;
            # this class should be derived from that class to inherit
            # this method; otherwise a simple function (not tied to a class)
            # should be used.
            ArchIFC.IfcRoot.setObjIfcComplexAttributeValue(
                self, self.object, "RepresentationContexts", data
            )
        except Exception:
            # This scenario occurs validly in IFC2X3,
            # as the mapConversion does not exist
            return

    def extractTargetCRSData(self, targetCRS):
        mappings = {
            "name": "Name",
            "description": "Description",
            "geodetic_datum": "GeodeticDatum",
            "vertical_datum": "VerticalDatum",
            "map_projection": "MapProjection",
            "map_zone": "MapZone",
        }
        data = {}
        for attributeName, ifcName in mappings.items():
            data[attributeName] = str(getattr(targetCRS, ifcName))

        if targetCRS.MapUnit.Prefix:
            data["map_unit"] = targetCRS.MapUnit.Prefix.title() + targetCRS.MapUnit.Name.lower()
        else:
            data["map_unit"] = targetCRS.MapUnit.Name.title()

        return data

    def extractMapConversionData(self, mapConversion):
        mappings = {
            "eastings": "Eastings",
            "northings": "Northings",
            "orthogonal_height": "OrthogonalHeight",
            "x_axis_abscissa": "XAxisAbscissa",
            "x_axis_ordinate": "XAxisOrdinate",
            "scale": "Scale",
        }
        data = {}
        for attributeName, ifcName in mappings.items():
            data[attributeName] = str(getattr(mapConversion, ifcName))

        data["true_north"] = str(
            self.calculateTrueNorthAngle(mapConversion.XAxisAbscissa, mapConversion.XAxisOrdinate)
        )
        return data

    def calculateTrueNorthAngle(self, x, y):
        return round(math.degrees(math.atan2(y, x)) - 90, 6)


def buildRelProductsAnnotations(ifcfile, root_element="IfcProduct"):
    """Build the products and annotations relation table."""
    products = ifcfile.by_type(root_element)

    annotations = ifcfile.by_type("IfcAnnotation")
    tp = []
    for product in products:
        if product.is_a("IfcGrid") and (product not in annotations):
            annotations.append(product)
        elif product not in annotations:
            tp.append(product)

    # remove any leftover annotations from products
    products = sorted(tp, key=lambda prod: prod.id())

    return products, annotations


def buildRelProductRepresentation(ifcfile):
    """Build the product/representations relation table."""
    if DEBUG_prod_repr:
        _msg(32 * "-")
        _msg("Product-representation table")

    prodrepr = dict()

    i = 1
    for p in ifcfile.by_type("IfcProduct"):
        if hasattr(p, "Representation") and p.Representation:
            if DEBUG_prod_repr:
                _msg("{}: {}, {}, '{}'".format(i, p.id(), p.is_a(), p.Name))

            for it in p.Representation.Representations:
                for it1 in it.Items:
                    prodrepr.setdefault(p.id(), []).append(it1.id())
                    if it1.is_a("IfcBooleanResult"):
                        prodrepr.setdefault(p.id(), []).append(it1.FirstOperand.id())
                    elif it.Items[0].is_a("IfcMappedItem"):
                        prodrepr.setdefault(p.id(), []).append(
                            it1.MappingSource.MappedRepresentation.id()
                        )
                        if it1.MappingSource.MappedRepresentation.is_a("IfcShapeRepresentation"):
                            for it2 in it1.MappingSource.MappedRepresentation.Items:
                                prodrepr.setdefault(p.id(), []).append(it2.id())
            i += 1
    return prodrepr


def buildRelAdditions(ifcfile):
    """Build the additions relation table."""
    additions = {}  # { host:[child,...], ... }

    for r in ifcfile.by_type("IfcRelContainedInSpatialStructure"):
        additions.setdefault(r.RelatingStructure.id(), []).extend(
            [e.id() for e in r.RelatedElements]
        )
    for r in ifcfile.by_type("IfcRelAggregates"):
        additions.setdefault(r.RelatingObject.id(), []).extend([e.id() for e in r.RelatedObjects])

    return additions


def buildRelGroups(ifcfile):
    """Build the groups relation table."""
    groups = {}  # { host:[child,...], ... }     # used in structural IFC

    for r in ifcfile.by_type("IfcRelAssignsToGroup"):
        groups.setdefault(r.RelatingGroup.id(), []).extend([e.id() for e in r.RelatedObjects])

    return groups


def buildRelSubtractions(ifcfile):
    """Build the subtractions relation table."""
    subtractions = []  # [ [opening,host], ... ]

    for r in ifcfile.by_type("IfcRelVoidsElement"):
        subtractions.append([r.RelatedOpeningElement.id(), r.RelatingBuildingElement.id()])

    return subtractions


def buildRelMattable(ifcfile):
    """Build the mattable relation table."""
    mattable = {}  # { objid:matid }

    for r in ifcfile.by_type("IfcRelAssociatesMaterial"):
        # the related object might not exist
        # https://forum.freecad.org/viewtopic.php?f=39&t=58607
        if r.RelatedObjects:
            for o in r.RelatedObjects:
                if r.RelatingMaterial.is_a("IfcMaterial"):
                    mattable[o.id()] = r.RelatingMaterial.id()
                elif r.RelatingMaterial.is_a("IfcMaterialLayer"):
                    mattable[o.id()] = r.RelatingMaterial.Material.id()
                elif r.RelatingMaterial.is_a("IfcMaterialLayerSet"):
                    mattable[o.id()] = r.RelatingMaterial.MaterialLayers[0].Material.id()
                elif r.RelatingMaterial.is_a("IfcMaterialLayerSetUsage"):
                    mattable[o.id()] = r.RelatingMaterial.ForLayerSet.MaterialLayers[
                        0
                    ].Material.id()

    return mattable


# Color relation tables.
# Products can have a color, materials can have a color,
# and products can have a material.
# Colors for material assigned to a product, and color of the product itself
# can be different
def buildRelColors(ifcfile, prodrepr):
    """Build the colors relation table.

    Returns all IfcStyledItem colors, material and product colors.

    Returns
    -------
    dict
        A dictionary with `{id: (r,g,b), ...}` values.
    """
    colors = {}  # { id:(r,g,b) }
    style_material_id = {}  # { style_entity_id: material_id) }

    style_color_rgb = {}  # { style_entity_id: (r,g,b) }
    for r in ifcfile.by_type("IfcStyledItem"):
        if r.Styles and r.Styles[0].is_a("IfcPresentationStyleAssignment"):
            for style1 in r.Styles[0].Styles:
                if style1.is_a("IfcSurfaceStyle"):
                    for style2 in style1.Styles:
                        if style2.is_a("IfcSurfaceStyleRendering"):
                            if style2.SurfaceColour:
                                c = style2.SurfaceColour
                                style_color_rgb[r.id()] = (c.Red, c.Green, c.Blue)

        # Nova
        # FIXME: style_entity_id = { style_entity_id: product_id } not material_id ???
        # see https://forum.freecad.org/viewtopic.php?f=39&t=37940&start=10#p329491
        # last code change in these color code https://github.com/FreeCAD/FreeCAD/commit/2d1f6ab1
        """
        if r.Item:
            # print(r.id())
            # print(r.Item)  # IfcRepresentationItem or IfcShapeRepresentation
            for p in prodrepr.keys():
                if r.Item.id() in prodrepr[p]:
                    style_material_id[r.id()] = p
                    # print(p)
                    # print(ifcfile[p])  # product
        """

    # A much faster version for Nova style_material_id with product_ids
    # no material colors, Nova ifc files often do not have materials at all
    for p in prodrepr.keys():
        # print("\n")
        # print(ifcfile[p])  # IfcProduct
        # print(ifcfile[p].Representation)  # IfcProductDefinitionShape
        # print(ifcfile[p].Representation.Representations[0])  # IfcShapeRepresentation
        # print(ifcfile[p].Representation.Representations[0].Items[0])  # IfcRepresentationItem
        # print(ifcfile[p].Representation.Representations[0].Items[0].StyledByItem[0])  # IfcStyledItem
        # print(ifcfile[p].Representation.Representations[0].Items[0].StyledByItem[0].id())
        # print(p)
        representation_item = ifcfile[p].Representation.Representations[0].Items[0]
        if hasattr(representation_item, "StyledByItem") and representation_item.StyledByItem:
            style_material_id[representation_item.StyledByItem[0].id()] = p

    # Allplan, ArchiCAD
    for m in ifcfile.by_type("IfcMaterialDefinitionRepresentation"):
        for it in m.Representations:
            if it.Items:
                style_material_id[it.Items[0].id()] = m.RepresentedMaterial.id()

    # create colors out of style_color_rgb and style_material_id
    for k in style_material_id:
        if k in style_color_rgb:
            colors[style_material_id[k]] = style_color_rgb[k]

    return colors


def buildRelProductColors(ifcfile, prodrepr):
    """Build the colors relation table from a product.

    Returns
    -------
    dict
        A dictionary with `{id: (r,g,b), ...}` values.
    """
    if DEBUG_prod_repr:
        _msg(32 * "-")
        _msg("Product-color table")

    colors = dict()
    i = 0

    for p in prodrepr.keys():
        # see method getColorFromProduct()
        # it is a method for the redundant code inside this loop
        # which can be used to get the color from a product directly

        # Representation item, see `IfcRepresentationItem` documentation.
        # All kinds of geometric or topological representation items
        # `IfcExtrudedAreaSolid`, `IfcMappedItem`, `IfcFacetedBrep`,
        # `IfcBooleanResult`, `IfcBooleanClippingResult`, etc.
        _body = ifcfile[p].Representation.Representations[0]
        repr_item = _body.Items[0]

        if DEBUG_prod_colors:
            _msg(
                "{}: {}, {}, '{}', rep_item {}".format(
                    i, ifcfile[p].id(), ifcfile[p].is_a(), ifcfile[p].Name, repr_item
                )
            )
        # Get the geometric representations which have a presentation style.
        # All representation items have the inverse attribute `StyledByItem`
        # for this.
        # There will be geometric representations which do not have
        # a presentation style so `StyledByItem` will be empty.
        if repr_item.StyledByItem:
            if DEBUG_prod_colors:
                _msg("  StyledByItem -> {}".format(repr_item.StyledByItem))
            # it has to be a `IfcStyledItem`, no check needed
            styled_item = repr_item.StyledByItem[0]

            # Write into colors table if a `IfcStyledItem` exists
            # for this product, write `None` if something goes wrong
            # or if the ifc file has errors and thus no valid color
            # is returned
            colors[p] = getColorFromStyledItem(styled_item)

        i += 1
    return colors


def buildRelMaterialColors(ifcfile, prodrepr):
    # not implemented
    pass


def getColorFromProduct(product):

    if product.Representation:
        for rep in product.Representation.Representations:
            for item in rep.Items:
                for style in item.StyledByItem:
                    color = getColorFromStyledItem(style)
                    if color:
                        return color


def getColorFromMaterial(material):

    if material.HasRepresentation:
        rep = material.HasRepresentation[0]
        if hasattr(rep, "Representations") and rep.Representations:
            rep = rep.Representations[0]
            if rep.is_a("IfcStyledRepresentation"):
                return getColorFromStyledItem(rep)
    return None


def color2colorRGB(color_data):

    if color_data is None:
        return None

    color_rgb = [
        int(round(color_data[0] * 255, 0)),
        int(round(color_data[1] * 255, 0)),
        int(round(color_data[2] * 255, 0)),
    ]  # int(159.99) would return 159 not 160, thus round

    return color_rgb


def getColorFromStyledItem(styled_item):
    """Get color from the IfcStyledItem.

    Returns
    -------
    float, float, float, int
        A tuple with the red, green, blue, and transparency values.
        If the `IfcStyledItem` is a `IfcDraughtingPreDefinedColour`
        the transparency is set to 0.
        The first three values range from 0 to 1.0, while the transparency
        varies from 0 to 100.

    None
        Return `None` if `styled_item` is not of type `'IfcStyledItem'`
        or if there is any other problem getting a color.
    """

    if styled_item.is_a("IfcStyledRepresentation"):
        styled_item = styled_item.Items[0]

    if not styled_item.is_a("IfcStyledItem"):
        return None

    rgb_color = None
    transparency = None
    col = None

    # The `IfcStyledItem` holds presentation style information for products,
    # either explicitly for an `IfcGeometricRepresentationItem` being part of
    # an `IfcShapeRepresentation` assigned to a product, or by assigning
    # presentation information to `IfcMaterial` being assigned
    # as other representation for a product.

    # In current IFC release (IFC2x3) only one presentation style
    # assignment shall be assigned.
    # In IFC4 `IfcPresentationStyleAssignment` is deprecated
    # In IFC4 multiple styles are assigned to style in 'IfcStyleItem' instead

    # print(ifcfile[p])
    # print(styled_item)
    # print(styled_item.Styles)
    if len(styled_item.Styles) == 0:
        # IN IFC2x3, only one element in `Styles` should be available.
        _wrn("No 'Style' in 'IfcStyleItem', do nothing.")
        # ca 100x in 210_King_Merged.ifc
        # Empty styles, #4952778=IfcStyledItem(#4952779,(),$)
        # this is an error in the IFC file in my opinion
    else:
        # never seen an ifc with more than one Styles in IfcStyledItem
        # the above seems to only apply for IFC2x3, IFC4 can have them
        # see https://forum.freecad.org/viewtopic.php?f=39&t=33560&p=437056#p437056

        # Get the `IfcPresentationStyleAssignment`, there should only be one,
        if styled_item.Styles[0].is_a("IfcPresentationStyleAssignment"):
            assign_style = styled_item.Styles[0]
        else:
            # `IfcPresentationStyleAssignment` is deprecated in IFC4,
            # in favor of `IfcStyleAssignmentSelect`
            assign_style = styled_item
        # print(assign_style)  # IfcPresentationStyleAssignment

        # `IfcPresentationStyleAssignment` can hold various kinds and counts
        # of styles, see `IfcPresentationStyleSelect`
        if assign_style.Styles[0].is_a("IfcSurfaceStyle"):
            _style = assign_style.Styles[0]
            # Schependomlaan and Nova and others
            # `IfcSurfaceStyleRendering`
            # print(_style.Styles[0])
            # `IfcColourRgb`
            rgb_color = _style.Styles[0].SurfaceColour
            # print(rgb_color)
            if (
                _style.Styles[0].is_a("IfcSurfaceStyleShading")
                and hasattr(_style.Styles[0], "Transparency")
                and _style.Styles[0].Transparency
            ):
                transparency = _style.Styles[0].Transparency * 100
        elif assign_style.Styles[0].is_a("IfcCurveStyle"):
            if len(assign_style.Styles) == 2 and assign_style.Styles[1].is_a("IfcSurfaceStyle"):
                # Allplan, new IFC export started in 2017
                # `IfcDraughtingPreDefinedColour`
                # print(assign_style.Styles[0].CurveColour)
                # TODO: check this; on index 1, is this what we need?!
                rgb_color = assign_style.Styles[1].Styles[0].SurfaceColour
                # print(rgb_color)
            else:
                # 2x Annotations in 210_King_Merged.ifc
                # print(ifcfile[p])
                # print(assign_style.Styles[0])
                # print(assign_style.Styles[0].CurveColour)
                rgb_color = assign_style.Styles[0].CurveColour

    if rgb_color:
        if rgb_color.is_a("IfcDraughtingPreDefinedColour"):
            if DEBUG_prod_colors:
                _msg("  '{}'= ".format(rgb_color.Name))

            col = predefined_to_rgb(rgb_color)

            if col:
                col = col + (0,)
        else:
            col = (
                rgb_color.Red,
                rgb_color.Green,
                rgb_color.Blue,
                int(transparency) if transparency else 0,
            )
    else:
        col = None

    if DEBUG_prod_colors:
        _msg("  {}".format(col))

    return col


def predefined_to_rgb(rgb_color):
    """Transform a predefined color name to its [r, g, b] representation.

    TODO: at the moment it doesn't handle 'by layer'.
    See: `IfcDraughtingPreDefinedColour` and `IfcPresentationLayerWithStyle`.
    """
    name = rgb_color.Name.lower()
    if name not in PREDEFINED_RGB:
        _wrn("Color name not in 'IfcDraughtingPreDefinedColour'.")

        if name == "by layer":
            _wrn(
                "'IfcDraughtingPreDefinedColour' set 'by layer'; "
                "currently not handled, set to 'None'."
            )
        return None

    return PREDEFINED_RGB[name]


# ************************************************************************************************
# property related methods


def buildRelProperties(ifcfile):
    """
    Builds and returns a dictionary of {object:[properties]} from an IFC file
    """

    # this method no longer used by the importer module
    # but this relation table might be useful anyway for other purposes

    properties = {}  # { objid : { psetid : [propertyid, ... ], ... }, ... }
    for r in ifcfile.by_type("IfcRelDefinesByProperties"):
        for obj in r.RelatedObjects:
            if not obj.id() in properties:
                properties[obj.id()] = {}
            psets = {}
            props = []
            if r.RelatingPropertyDefinition.is_a("IfcPropertySet"):
                props.extend([prop.id() for prop in r.RelatingPropertyDefinition.HasProperties])
                psets[r.RelatingPropertyDefinition.id()] = props
                properties[obj.id()].update(psets)
    return properties


def getIfcPropertySets(ifcfile, pid):
    """Returns a dictionary of {pset_id:[prop_id, prop_id...]} for an IFC object"""

    # get psets for this pid
    psets = {}
    for rel in ifcfile[pid].IsDefinedBy:
        # the following if condition is needed in IFC2x3 only
        # https://forum.freecad.org/viewtopic.php?f=39&t=37892#p322884
        if rel.is_a("IfcRelDefinesByProperties"):
            props = []
            if rel.RelatingPropertyDefinition.is_a("IfcPropertySet"):
                props.extend([prop.id() for prop in rel.RelatingPropertyDefinition.HasProperties])
                psets[rel.RelatingPropertyDefinition.id()] = props
    return psets


def getIfcProperties(ifcfile, pid, psets, d):
    """builds valid property values for FreeCAD"""

    for pset in psets.keys():
        # print("reading pset: ",pset)
        psetname = ifcfile[pset].Name
        for prop in psets[pset]:
            e = ifcfile[prop]
            pname = e.Name
            if e.is_a("IfcPropertySingleValue"):
                if e.NominalValue:
                    ptype = e.NominalValue.is_a()
                    if ptype in ["IfcLabel", "IfcText", "IfcIdentifier", "IfcDescriptiveMeasure"]:
                        pvalue = e.NominalValue.wrappedValue
                    else:
                        pvalue = str(e.NominalValue.wrappedValue)
                    if hasattr(e.NominalValue, "Unit"):
                        if e.NominalValue.Unit:
                            pvalue += e.NominalValue.Unit
                    d[pname + ";;" + psetname] = ptype + ";;" + pvalue
                # print("adding property: ",pname,ptype,pvalue," pset ",psetname)
    return d


def getIfcPsetProperties(ifcfile, pid):
    """directly build the property table from pid and ifcfile for FreeCAD"""

    return getIfcProperties(ifcfile, pid, getIfcPropertySets(ifcfile, pid), {})


def getUnit(unit):
    """Get the unit multiplier for different decimal prefixes.

    Only for when the unit is METRE.
    When no Prefix is provided, return 1000, that is, mm x 1000 = metre.
    For other cases, return 1.0.
    """
    if unit.Name == "METRE":
        if unit.Prefix == "KILO":
            return 1000000.0
        elif unit.Prefix == "HECTO":
            return 100000.0
        elif unit.Prefix == "DECA":
            return 10000.0
        elif not unit.Prefix:
            return 1000.0
        elif unit.Prefix == "DECI":
            return 100.0
        elif unit.Prefix == "CENTI":
            return 10.0
    return 1.0


def getScaling(ifcfile):
    """Return a scaling factor from the IFC file; units to mm."""
    ua = ifcfile.by_type("IfcUnitAssignment")

    if not ua:
        return 1.0

    ua = ua[0]
    for u in ua.Units:
        if u.UnitType == "LENGTHUNIT":
            if u.is_a("IfcConversionBasedUnit"):
                f = getUnit(u.ConversionFactor.UnitComponent)
                return f * u.ConversionFactor.ValueComponent.wrappedValue
            elif u.is_a("IfcSIUnit") or u.is_a("IfcUnit"):
                return getUnit(u)
    return 1.0


def getRotation(entity):
    """returns a FreeCAD rotation from an IfcProduct with a IfcMappedItem representation"""
    try:
        u = FreeCAD.Vector(entity.Axis1.DirectionRatios)
        v = FreeCAD.Vector(entity.Axis2.DirectionRatios)
        w = FreeCAD.Vector(entity.Axis3.DirectionRatios)
    except AttributeError:
        return FreeCAD.Rotation()
    return FreeCAD.Rotation(u, v, w, "ZYX")


def getPlacement(entity, scaling=1000):
    """returns a placement from the given entity"""

    if not entity:
        return None
    import DraftVecUtils

    pl = None
    if entity.is_a("IfcAxis2Placement3D"):
        x = getVector(entity.RefDirection, scaling)
        z = getVector(entity.Axis, scaling)
        if x and z:
            y = z.cross(x)
            m = DraftVecUtils.getPlaneRotation(x, y, z)
            pl = FreeCAD.Placement(m)
        else:
            pl = FreeCAD.Placement()
        loc = getVector(entity.Location, scaling)
        if loc:
            pl.move(loc)
    elif entity.is_a("IfcAxis2Placement2D"):
        _wrn("not implemented IfcAxis2Placement2D, ", end="")
    elif entity.is_a("IfcLocalPlacement"):
        pl = getPlacement(entity.PlacementRelTo, 1)  # original placement
        relpl = getPlacement(entity.RelativePlacement, 1)  # relative transf
        if pl and relpl:
            pl = pl.multiply(relpl)
        elif relpl:
            pl = relpl
    elif entity.is_a("IfcCartesianPoint"):
        loc = getVector(entity, scaling)
        pl = FreeCAD.Placement()
        pl.move(loc)
    if pl:
        pl.Base = FreeCAD.Vector(pl.Base).multiply(scaling)
    return pl


def getVector(entity, scaling=1000):
    """returns a vector from the given entity"""

    if not entity:
        return None
    v = None
    if entity.is_a("IfcDirection"):
        if len(entity.DirectionRatios) == 3:
            v = FreeCAD.Vector(tuple(entity.DirectionRatios))
        else:
            v = FreeCAD.Vector(tuple(entity.DirectionRatios + [0]))
    elif entity.is_a("IfcCartesianPoint"):
        if len(entity.Coordinates) == 3:
            v = FreeCAD.Vector(tuple(entity.Coordinates))
        else:
            v = FreeCAD.Vector(tuple(entity.Coordinates + [0]))
    # if v:
    #     v.multiply(scaling)
    return v


def get2DShape(representation, scaling=1000, notext=False):
    """Returns a shape from a 2D IfcShapeRepresentation
    if notext is True, no Draft text is created"""

    import Part
    import DraftVecUtils
    import Draft

    def getPolyline(ent):
        pts = []
        for p in ent.Points:
            c = p.Coordinates
            c = FreeCAD.Vector(c[0], c[1], c[2] if len(c) > 2 else 0)
            c.multiply(scaling)
            pts.append(c)
        return Part.makePolygon(pts)

    def getRectangle(ent):
        return Part.makePlane(ent.XDim, ent.YDim)

    def getLine(ent):
        pts = []
        p1 = getVector(ent.Pnt)
        p1.multiply(scaling)
        pts.append(p1)
        p2 = getVector(ent.Dir)
        p2.multiply(scaling)
        p2 = p1.add(p2)
        pts.append(p2)
        return Part.makePolygon(pts)

    def getCircle(ent):
        c = ent.Position.Location.Coordinates
        c = FreeCAD.Vector(c[0], c[1], c[2] if len(c) > 2 else 0)
        c.multiply(scaling)
        r = ent.Radius * scaling
        return Part.makeCircle(r, c)

    def getCurveSet(ent):
        result = []
        if ent.is_a() in ["IfcGeometricCurveSet", "IfcGeometricSet"]:
            elts = ent.Elements
        elif ent.is_a() in [
            "IfcLine",
            "IfcPolyline",
            "IfcCircle",
            "IfcTrimmedCurve",
            "IfcRectangleProfileDef",
        ]:
            elts = [ent]
        else:
            print("getCurveSet: unhandled entity: ", ent)
            return []

        for el in elts:
            if el.is_a("IfcPolyline"):
                result.append(getPolyline(el))
            elif el.is_a("IfcRectangleProfileDef"):
                result.append(getRectangle(el))
            elif el.is_a("IfcLine"):
                result.append(getLine(el))
            elif el.is_a("IfcCircle"):
                result.append(getCircle(el))
            elif el.is_a("IfcTrimmedCurve"):
                base = el.BasisCurve
                t1 = el.Trim1[0].wrappedValue
                t2 = el.Trim2[0].wrappedValue
                if not el.SenseAgreement:
                    t1, t2 = t2, t1
                if base.is_a("IfcPolyline"):
                    bc = getPolyline(base)
                    result.append(bc)
                elif base.is_a("IfcCircle"):
                    bc = getCircle(base)
                    e = Part.ArcOfCircle(bc.Curve, math.radians(t1), math.radians(t2)).toShape()
                    d = base.Position.RefDirection.DirectionRatios
                    v = FreeCAD.Vector(d[0], d[1], d[2] if len(d) > 2 else 0)
                    a = -DraftVecUtils.angle(v)
                    e.rotate(bc.Curve.Center, FreeCAD.Vector(0, 0, 1), math.degrees(a))
                    result.append(e)
            elif el.is_a("IfcCompositeCurve"):
                for base in el.Segments:
                    if base.ParentCurve.is_a("IfcPolyline"):
                        bc = getPolyline(base.ParentCurve)
                        result.append(bc)
                    elif base.ParentCurve.is_a("IfcCircle"):
                        bc = getCircle(base.ParentCurve)
                        e = Part.ArcOfCircle(bc.Curve, math.radians(t1), math.radians(t2)).toShape()
                        d = base.Position.RefDirection.DirectionRatios
                        v = FreeCAD.Vector(d[0], d[1], d[2] if len(d) > 2 else 0)
                        a = -DraftVecUtils.angle(v)
                        e.rotate(bc.Curve.Center, FreeCAD.Vector(0, 0, 1), math.degrees(a))
                        result.append(e)
            elif el.is_a("IfcIndexedPolyCurve"):
                coords = el.Points.CoordList

                def index2points(segment):
                    pts = []
                    for i in segment.wrappedValue:
                        c = coords[i - 1]
                        c = FreeCAD.Vector(c[0], c[1], c[2] if len(c) > 2 else 0)
                        c.multiply(scaling)
                        pts.append(c)
                    return pts

                if not el.Segments:
                    # use all points
                    verts = [FreeCAD.Vector(c[0], c[1], c[2] if len(c) > 2 else 0) for c in coords]
                    verts = [v.multiply(scaling) for v in verts]
                    result.append(Part.makePolygon(verts))
                else:
                    for s in el.Segments:
                        if s.is_a("IfcLineIndex"):
                            result.append(Part.makePolygon(index2points(s)))
                        elif s.is_a("IfcArcIndex"):
                            [p1, p2, p3] = index2points(s)
                            result.append(Part.Arc(p1, p2, p3))
                        else:
                            raise RuntimeError("Illegal IfcIndexedPolyCurve segment: " + s.is_a())
            else:
                print("importIFCHelper.getCurveSet: unhandled element: ", el)

        return result

    result = []
    if representation.is_a("IfcShapeRepresentation"):
        for item in representation.Items:
            if item.is_a() in ["IfcGeometricCurveSet", "IfcGeometricSet"]:
                result = getCurveSet(item)
            elif item.is_a("IfcMappedItem"):
                preresult = get2DShape(item.MappingSource.MappedRepresentation, scaling)
                pla = getPlacement(item.MappingSource.MappingOrigin, scaling)
                rot = getRotation(item.MappingTarget)
                if pla:
                    if rot.Angle:
                        pla.Rotation = rot
                    for r in preresult:
                        # r.Placement = pla
                        result.append(r)
                else:
                    result = preresult
            elif item.is_a("IfcTextLiteral"):
                if notext:
                    continue
                pl = getPlacement(item.Placement, scaling)
                if pl:
                    t = Draft.make_text(item.Literal.split(";"), pl)
                    if FreeCAD.GuiUp:
                        if item.Path == "RIGHT":
                            t.ViewObject.Justification = "Right"
                    # do not return because there might be more than one representation
                    # return []  # TODO dirty hack... Object creation should not be done here
    elif representation.is_a() in [
        "IfcPolyline",
        "IfcCircle",
        "IfcTrimmedCurve",
        "IfcRectangleProfileDef",
    ]:
        result = getCurveSet(representation)
    return result


def getProfileCenterPoint(sweptsolid):
    """returns the center point of the profile of an extrusion"""
    v = FreeCAD.Vector(0, 0, 0)
    if hasattr(sweptsolid, "SweptArea"):
        profile = get2DShape(sweptsolid.SweptArea)
        if profile:
            profile = profile[0]
            if hasattr(profile, "CenterOfMass"):
                v = profile.CenterOfMass
            elif hasattr(profile, "BoundBox"):
                v = profile.BoundBox.Center
    if hasattr(sweptsolid, "Position"):
        pos = getPlacement(sweptsolid.Position)
        v = pos.multVec(v)
    return v


def isRectangle(verts):
    """returns True if the given 4 vertices form a rectangle"""
    if len(verts) != 4:
        return False
    v1 = verts[1].sub(verts[0])
    v2 = verts[2].sub(verts[1])
    v3 = verts[3].sub(verts[2])
    v4 = verts[0].sub(verts[3])
    if abs(v2.getAngle(v1) - math.pi / 2) > 0.01:
        return False
    if abs(v3.getAngle(v2) - math.pi / 2) > 0.01:
        return False
    if abs(v4.getAngle(v3) - math.pi / 2) > 0.01:
        return False
    return True


def createFromProperties(propsets, ifcfile, parametrics):
    """
    Creates a FreeCAD parametric object from a set of properties.
    """

    obj = None
    sets = []
    appset = None
    guiset = None
    for pset in propsets.keys():
        if ifcfile[pset].Name == "FreeCADPropertySet":
            appset = {}
            for pid in propsets[pset]:
                p = ifcfile[pid]
                appset[p.Name] = p.NominalValue.wrappedValue
        elif ifcfile[pset].Name == "FreeCADGuiPropertySet":
            guiset = {}
            for pid in propsets[pset]:
                p = ifcfile[pid]
                guiset[p.Name] = p.NominalValue.wrappedValue
    if appset:
        oname = None
        otype = None
        if "FreeCADType" in appset:
            if "FreeCADName" in appset:
                obj = FreeCAD.ActiveDocument.addObject(appset["FreeCADType"], appset["FreeCADName"])
                if "FreeCADAppObject" in appset:
                    mod, cla = appset["FreeCADAppObject"].split(".")
                    if "'" in mod:
                        mod = mod.split("'")[-1]
                    if "'" in cla:
                        cla = cla.split("'")[0]
                    import importlib

                    mod = importlib.import_module(mod)
                    getattr(mod, cla)(obj)
                sets.append(("App", appset))
                if FreeCAD.GuiUp:
                    if guiset:
                        if "FreeCADGuiObject" in guiset:
                            mod, cla = guiset["FreeCADGuiObject"].split(".")
                            if "'" in mod:
                                mod = mod.split("'")[-1]
                            if "'" in cla:
                                cla = cla.split("'")[0]
                            import importlib

                            mod = importlib.import_module(mod)
                            getattr(mod, cla)(obj.ViewObject)
                        sets.append(("Gui", guiset))
    if obj and sets:
        for realm, pset in sets:
            if realm == "App":
                target = obj
            else:
                target = obj.ViewObject
            for key, val in pset.items():
                if key.startswith("FreeCAD_") or key.startswith("FreeCADGui_"):
                    name = key.split("_")[1]
                    if name in target.PropertiesList:
                        if not target.getEditorMode(name):
                            ptype = target.getTypeIdOfProperty(name)
                            if ptype in [
                                "App::PropertyString",
                                "App::PropertyEnumeration",
                                "App::PropertyInteger",
                                "App::PropertyFloat",
                            ]:
                                setattr(target, name, val)
                            elif ptype in ["App::PropertyLength", "App::PropertyDistance"]:
                                setattr(target, name, val * 1000)
                            elif ptype == "App::PropertyBool":
                                if val in [".T.", True]:
                                    setattr(target, name, True)
                                else:
                                    setattr(target, name, False)
                            elif ptype == "App::PropertyVector":
                                setattr(
                                    target,
                                    name,
                                    FreeCAD.Vector(
                                        [float(s) for s in val.split("(")[1].strip(")").split(",")]
                                    ),
                                )
                            elif ptype == "App::PropertyArea":
                                setattr(target, name, val * 1000000)
                            elif ptype == "App::PropertyPlacement":
                                data = val.split("[")[1].strip("]").split("(")
                                data = [data[1].split(")")[0], data[2].strip(")")]
                                v = FreeCAD.Vector([float(s) for s in data[0].split(",")])
                                r = FreeCAD.Rotation(*[float(s) for s in data[1].split(",")])
                                setattr(target, name, FreeCAD.Placement(v, r))
                            elif ptype == "App::PropertyLink":
                                link = val.split("_")[1]
                                parametrics.append([target, name, link])
                            else:
                                print("Unhandled FreeCAD property:", name, " of type:", ptype)
    return obj, parametrics


def applyColorDict(doc, colordict=None):
    """applies the contents of a color dict to the objects in the given doc.
    If no colordict is given, the doc Meta property is searched for a "colordict" entry."""

    if not colordict:
        if "colordict" in doc.Meta:
            import json

            colordict = json.loads(doc.Meta["colordict"])
    if colordict:
        for obj in doc.Objects:
            if obj.Name in colordict:
                color = colordict[obj.Name]
                if hasattr(obj.ViewObject, "ShapeColor"):
                    obj.ViewObject.ShapeColor = tuple(color[0:3])
                if hasattr(obj.ViewObject, "Transparency") and (len(color) >= 4):
                    obj.ViewObject.Transparency = 1.0 - color[3]
    else:
        print("No valid color dict to apply")


def getParents(ifcobj):
    """finds the parent entities of an IFC entity"""

    parentlist = []
    if hasattr(ifcobj, "ContainedInStructure"):
        for rel in ifcobj.ContainedInStructure:
            parentlist.append(rel.RelatingStructure)
    elif hasattr(ifcobj, "Decomposes"):
        for rel in ifcobj.Decomposes:
            if rel.is_a("IfcRelAggregates"):
                parentlist.append(rel.RelatingObject)
    return parentlist


def createAnnotation(annotation, doc, ifcscale, preferences):
    """creates an annotation object"""

    anno = None
    aid = annotation.id()
    if annotation.is_a("IfcGrid"):
        axes = []
        uvwaxes = ()
        if annotation.UAxes:
            uvwaxes = annotation.UAxes
        if annotation.VAxes:
            uvwaxes = uvwaxes + annotation.VAxes
        if annotation.WAxes:
            uvwaxes = uvwaxes + annotation.WAxes
        for axis in uvwaxes:
            if axis.AxisCurve:
                sh = get2DShape(axis.AxisCurve, ifcscale)
                if sh and (len(sh[0].Vertexes) == 2):  # currently only straight axes are supported
                    sh = sh[0]
                    l = sh.Length
                    pl = FreeCAD.Placement()
                    pl.Base = sh.Vertexes[0].Point
                    pl.Rotation = FreeCAD.Rotation(
                        FreeCAD.Vector(0, 1, 0), sh.Vertexes[-1].Point.sub(sh.Vertexes[0].Point)
                    )
                    o = Arch.makeAxis(1, l)
                    o.Length = l
                    o.Placement = pl
                    o.CustomNumber = axis.AxisTag
                    axes.append(o)
        if axes:
            name = "Grid"
            grid_placement = None
            if annotation.Name:
                name = annotation.Name
            if annotation.ObjectPlacement:
                # https://forum.freecad.org/viewtopic.php?f=39&t=40027
                grid_placement = getPlacement(annotation.ObjectPlacement, scaling=1)
            if preferences["PREFIX_NUMBERS"]:
                name = "ID" + str(aid) + " " + name
            anno = Arch.makeAxisSystem(axes, name)
            if grid_placement:
                anno.Placement = grid_placement
        print(" axis")
    else:
        name = "Annotation"
        if annotation.Name:
            name = annotation.Name
        if "annotation" not in name.lower():
            name = "Annotation " + name
        if preferences["PREFIX_NUMBERS"]:
            name = "ID" + str(aid) + " " + name
        shapes2d = []
        for rep in annotation.Representation.Representations:
            if rep.RepresentationIdentifier in ["Annotation", "FootPrint", "Axis"]:
                sh = get2DShape(rep, ifcscale)
                if sh in doc.Objects:
                    # dirty hack: get2DShape might return an object directly if non-shape based (texts for ex)
                    anno = sh
                else:
                    shapes2d.extend(sh)
        if shapes2d:
            import Part

            sh = Part.makeCompound(shapes2d)
            # if preferences['DEBUG']: print(" shape")
            anno = doc.addObject("Part::Feature", name)
            anno.Shape = sh
            p = getPlacement(annotation.ObjectPlacement, ifcscale)
            if p:  # and annotation.is_a("IfcAnnotation"):
                anno.Placement = p
        # else:
        # if preferences['DEBUG']: print(" no shape")

    return anno



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Vector as vec

from BOPTools import SplitFeatures
from Part import makeLine

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Truss 3D circle cs, seg3 elements",
        "meshtype": "edge",
        "meshelement": "Seg3",
        "constraints": ["fixed", "force"],
        "solvers": ["ccxtools"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

# To run the example from Python console use,
# (works even after an edit without restart of FreeCAD):
import importlib
import femexamples.truss_3d_cs_circle_ele_seg3 as example
importlib.reload(example)
example.setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=23318&start=20#p521050
https://forum.freecad.org/viewtopic.php?f=13&t=59239&start=100#p521220

Z88 official example 2, crane beam
- nonsway (on x deformation on both support)
- in practical engineering one would let the truss breathe on one side
- max deflection CalxuliX: 8.22 mm  # but it needs four seg3 elements foreach bar
- max deflection Mystran : x.xx mm
- max deflection Z88 : 8.19 mm  # one seg2 truss element foreach bar
"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric object
    # load line
    load_line = doc.addObject("Part::Line", "LoadLine")
    load_line.X1 = 0
    load_line.Y1 = 0
    load_line.Z1 = 1000
    load_line.X2 = 0
    load_line.Y2 = 0
    load_line.Z2 = 0
    if FreeCAD.GuiUp:
        load_line.ViewObject.hide()

    # commands were generated by Python out of the original Z88 Mesh obj data
    v1 = vec(0.0, 2000.0, 0.0)
    v2 = vec(0.0, 0.0, 0.0)
    v3 = vec(1000.0, 1000.0, 2000.0)
    v4 = vec(2000.0, 2000.0, 0.0)
    v5 = vec(2000.0, 0.0, 0.0)
    v6 = vec(3000.0, 1000.0, 2000.0)
    v7 = vec(4000.0, 2000.0, 0.0)
    v8 = vec(4000.0, 0.0, 0.0)
    v9 = vec(5000.0, 1000.0, 2000.0)
    v10 = vec(6000.0, 2000.0, 0.0)
    v11 = vec(6000.0, 0.0, 0.0)
    v12 = vec(7000.0, 1000.0, 2000.0)
    v13 = vec(8000.0, 2000.0, 0.0)
    v14 = vec(8000.0, 0.0, 0.0)
    v15 = vec(9000.0, 1000.0, 2000.0)
    v16 = vec(10000.0, 2000.0, 0.0)
    v17 = vec(10000.0, 0.0, 0.0)
    v18 = vec(11000.0, 1000.0, 2000.0)
    v19 = vec(12000.0, 2000.0, 0.0)
    v20 = vec(12000.0, 0.0, 0.0)
    line1 = makeLine(v1, v2)
    line2 = makeLine(v4, v5)
    line3 = makeLine(v7, v8)
    line4 = makeLine(v10, v11)
    line5 = makeLine(v13, v14)
    line6 = makeLine(v16, v17)
    line7 = makeLine(v19, v20)
    line8 = makeLine(v1, v4)
    line9 = makeLine(v2, v5)
    line10 = makeLine(v4, v7)
    line11 = makeLine(v5, v8)
    line12 = makeLine(v7, v10)
    line13 = makeLine(v8, v11)
    line14 = makeLine(v10, v13)
    line15 = makeLine(v11, v14)
    line16 = makeLine(v13, v16)
    line17 = makeLine(v14, v17)
    line18 = makeLine(v16, v19)
    line19 = makeLine(v17, v20)
    line20 = makeLine(v1, v3)
    line21 = makeLine(v4, v6)
    line22 = makeLine(v7, v9)
    line23 = makeLine(v10, v12)
    line24 = makeLine(v13, v15)
    line25 = makeLine(v16, v18)
    line26 = makeLine(v2, v3)
    line27 = makeLine(v5, v6)
    line28 = makeLine(v8, v9)
    line29 = makeLine(v11, v12)
    line30 = makeLine(v14, v15)
    line31 = makeLine(v17, v18)
    line32 = makeLine(v3, v4)
    line33 = makeLine(v6, v7)
    line34 = makeLine(v9, v10)
    line35 = makeLine(v12, v13)
    line36 = makeLine(v15, v16)
    line37 = makeLine(v18, v19)
    line38 = makeLine(v3, v5)
    line39 = makeLine(v6, v8)
    line40 = makeLine(v9, v11)
    line41 = makeLine(v12, v14)
    line42 = makeLine(v15, v17)
    line43 = makeLine(v18, v20)
    line44 = makeLine(v3, v6)
    line45 = makeLine(v6, v9)
    line46 = makeLine(v9, v12)
    line47 = makeLine(v12, v15)
    line48 = makeLine(v15, v18)
    line49 = makeLine(v2, v4)
    line50 = makeLine(v5, v7)
    line51 = makeLine(v8, v10)
    line52 = makeLine(v11, v13)
    line53 = makeLine(v14, v16)
    line54 = makeLine(v17, v19)
    obj_line1 = doc.addObject("Part::Feature", "Line1")
    obj_line1.Shape = line1
    obj_line2 = doc.addObject("Part::Feature", "Line2")
    obj_line2.Shape = line2
    obj_line3 = doc.addObject("Part::Feature", "Line3")
    obj_line3.Shape = line3
    obj_line4 = doc.addObject("Part::Feature", "Line4")
    obj_line4.Shape = line4
    obj_line5 = doc.addObject("Part::Feature", "Line5")
    obj_line5.Shape = line5
    obj_line6 = doc.addObject("Part::Feature", "Line6")
    obj_line6.Shape = line6
    obj_line7 = doc.addObject("Part::Feature", "Line7")
    obj_line7.Shape = line7
    obj_line8 = doc.addObject("Part::Feature", "Line8")
    obj_line8.Shape = line8
    obj_line9 = doc.addObject("Part::Feature", "Line9")
    obj_line9.Shape = line9
    obj_line10 = doc.addObject("Part::Feature", "Line10")
    obj_line10.Shape = line10
    obj_line11 = doc.addObject("Part::Feature", "Line11")
    obj_line11.Shape = line11
    obj_line12 = doc.addObject("Part::Feature", "Line12")
    obj_line12.Shape = line12
    obj_line13 = doc.addObject("Part::Feature", "Line13")
    obj_line13.Shape = line13
    obj_line14 = doc.addObject("Part::Feature", "Line14")
    obj_line14.Shape = line14
    obj_line15 = doc.addObject("Part::Feature", "Line15")
    obj_line15.Shape = line15
    obj_line16 = doc.addObject("Part::Feature", "Line16")
    obj_line16.Shape = line16
    obj_line17 = doc.addObject("Part::Feature", "Line17")
    obj_line17.Shape = line17
    obj_line18 = doc.addObject("Part::Feature", "Line18")
    obj_line18.Shape = line18
    obj_line19 = doc.addObject("Part::Feature", "Line19")
    obj_line19.Shape = line19
    obj_line20 = doc.addObject("Part::Feature", "Line20")
    obj_line20.Shape = line20
    obj_line21 = doc.addObject("Part::Feature", "Line21")
    obj_line21.Shape = line21
    obj_line22 = doc.addObject("Part::Feature", "Line22")
    obj_line22.Shape = line22
    obj_line23 = doc.addObject("Part::Feature", "Line23")
    obj_line23.Shape = line23
    obj_line24 = doc.addObject("Part::Feature", "Line24")
    obj_line24.Shape = line24
    obj_line25 = doc.addObject("Part::Feature", "Line25")
    obj_line25.Shape = line25
    obj_line26 = doc.addObject("Part::Feature", "Line26")
    obj_line26.Shape = line26
    obj_line27 = doc.addObject("Part::Feature", "Line27")
    obj_line27.Shape = line27
    obj_line28 = doc.addObject("Part::Feature", "Line28")
    obj_line28.Shape = line28
    obj_line29 = doc.addObject("Part::Feature", "Line29")
    obj_line29.Shape = line29
    obj_line30 = doc.addObject("Part::Feature", "Line30")
    obj_line30.Shape = line30
    obj_line31 = doc.addObject("Part::Feature", "Line31")
    obj_line31.Shape = line31
    obj_line32 = doc.addObject("Part::Feature", "Line32")
    obj_line32.Shape = line32
    obj_line33 = doc.addObject("Part::Feature", "Line33")
    obj_line33.Shape = line33
    obj_line34 = doc.addObject("Part::Feature", "Line34")
    obj_line34.Shape = line34
    obj_line35 = doc.addObject("Part::Feature", "Line35")
    obj_line35.Shape = line35
    obj_line36 = doc.addObject("Part::Feature", "Line36")
    obj_line36.Shape = line36
    obj_line37 = doc.addObject("Part::Feature", "Line37")
    obj_line37.Shape = line37
    obj_line38 = doc.addObject("Part::Feature", "Line38")
    obj_line38.Shape = line38
    obj_line39 = doc.addObject("Part::Feature", "Line39")
    obj_line39.Shape = line39
    obj_line40 = doc.addObject("Part::Feature", "Line40")
    obj_line40.Shape = line40
    obj_line41 = doc.addObject("Part::Feature", "Line41")
    obj_line41.Shape = line41
    obj_line42 = doc.addObject("Part::Feature", "Line42")
    obj_line42.Shape = line42
    obj_line43 = doc.addObject("Part::Feature", "Line43")
    obj_line43.Shape = line43
    obj_line44 = doc.addObject("Part::Feature", "Line44")
    obj_line44.Shape = line44
    obj_line45 = doc.addObject("Part::Feature", "Line45")
    obj_line45.Shape = line45
    obj_line46 = doc.addObject("Part::Feature", "Line46")
    obj_line46.Shape = line46
    obj_line47 = doc.addObject("Part::Feature", "Line47")
    obj_line47.Shape = line47
    obj_line48 = doc.addObject("Part::Feature", "Line48")
    obj_line48.Shape = line48
    obj_line49 = doc.addObject("Part::Feature", "Line49")
    obj_line49.Shape = line49
    obj_line50 = doc.addObject("Part::Feature", "Line50")
    obj_line50.Shape = line50
    obj_line51 = doc.addObject("Part::Feature", "Line51")
    obj_line51.Shape = line51
    obj_line52 = doc.addObject("Part::Feature", "Line52")
    obj_line52.Shape = line52
    obj_line53 = doc.addObject("Part::Feature", "Line53")
    obj_line53.Shape = line53
    obj_line54 = doc.addObject("Part::Feature", "Line54")
    obj_line54.Shape = line54
    doc.recompute()
    geom_obj = SplitFeatures.makeBooleanFragments(name="CraneTruss")
    geom_obj.Objects = [
        obj_line1,
        obj_line2,
        obj_line3,
        obj_line4,
        obj_line5,
        obj_line6,
        obj_line7,
        obj_line8,
        obj_line9,
        obj_line10,
        obj_line11,
        obj_line12,
        obj_line13,
        obj_line14,
        obj_line15,
        obj_line16,
        obj_line17,
        obj_line18,
        obj_line19,
        obj_line20,
        obj_line21,
        obj_line22,
        obj_line23,
        obj_line24,
        obj_line25,
        obj_line26,
        obj_line27,
        obj_line28,
        obj_line29,
        obj_line30,
        obj_line31,
        obj_line32,
        obj_line33,
        obj_line34,
        obj_line35,
        obj_line36,
        obj_line37,
        obj_line38,
        obj_line39,
        obj_line40,
        obj_line41,
        obj_line42,
        obj_line43,
        obj_line44,
        obj_line45,
        obj_line46,
        obj_line47,
        obj_line48,
        obj_line49,
        obj_line50,
        obj_line51,
        obj_line52,
        obj_line53,
        obj_line54,
    ]
    if FreeCAD.GuiUp:
        obj_line1.ViewObject.hide()
        obj_line2.ViewObject.hide()
        obj_line3.ViewObject.hide()
        obj_line4.ViewObject.hide()
        obj_line5.ViewObject.hide()
        obj_line6.ViewObject.hide()
        obj_line7.ViewObject.hide()
        obj_line8.ViewObject.hide()
        obj_line9.ViewObject.hide()
        obj_line10.ViewObject.hide()
        obj_line11.ViewObject.hide()
        obj_line12.ViewObject.hide()
        obj_line13.ViewObject.hide()
        obj_line14.ViewObject.hide()
        obj_line15.ViewObject.hide()
        obj_line16.ViewObject.hide()
        obj_line17.ViewObject.hide()
        obj_line18.ViewObject.hide()
        obj_line19.ViewObject.hide()
        obj_line20.ViewObject.hide()
        obj_line21.ViewObject.hide()
        obj_line22.ViewObject.hide()
        obj_line23.ViewObject.hide()
        obj_line24.ViewObject.hide()
        obj_line25.ViewObject.hide()
        obj_line26.ViewObject.hide()
        obj_line27.ViewObject.hide()
        obj_line28.ViewObject.hide()
        obj_line29.ViewObject.hide()
        obj_line30.ViewObject.hide()
        obj_line31.ViewObject.hide()
        obj_line32.ViewObject.hide()
        obj_line33.ViewObject.hide()
        obj_line34.ViewObject.hide()
        obj_line35.ViewObject.hide()
        obj_line36.ViewObject.hide()
        obj_line37.ViewObject.hide()
        obj_line38.ViewObject.hide()
        obj_line39.ViewObject.hide()
        obj_line40.ViewObject.hide()
        obj_line41.ViewObject.hide()
        obj_line42.ViewObject.hide()
        obj_line43.ViewObject.hide()
        obj_line44.ViewObject.hide()
        obj_line45.ViewObject.hide()
        obj_line46.ViewObject.hide()
        obj_line47.ViewObject.hide()
        obj_line48.ViewObject.hide()
        obj_line49.ViewObject.hide()
        obj_line50.ViewObject.hide()
        obj_line51.ViewObject.hide()
        obj_line52.ViewObject.hide()
        obj_line53.ViewObject.hide()
        obj_line54.ViewObject.hide()

    doc.recompute()
    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    elif solvertype == "z88":
        solver_obj = ObjectsFem.makeSolverZ88(doc, "SolverZ88")
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.SplitInputWriter = False
        solver_obj.AnalysisType = "static"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
    analysis.addObject(solver_obj)

    # beam section
    beamsection_obj = ObjectsFem.makeElementGeometry1D(
        doc, sectiontype="Circular", height=25.0, name="CrossSectionCircular"
    )
    analysis.addObject(beamsection_obj)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "MechanicalMaterial")
    mat = material_obj.Material
    mat["Name"] = "Steel"
    mat["YoungsModulus"] = "200000 MPa"
    mat["PoissonRatio"] = "0.30"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [(geom_obj, ("Vertex1", "Vertex2", "Vertex13", "Vertex14"))]
    analysis.addObject(con_fixed)

    # constraint force
    con_force = ObjectsFem.makeConstraintForce(doc, "ConstraintForce")
    con_force.References = [(geom_obj, ("Vertex5", "Vertex6"))]
    con_force.Force = "60000.0 N"  # 30 kN on each Node
    con_force.Direction = (load_line, ["Edge1"])
    con_force.Reversed = False
    analysis.addObject(con_force)

    # mesh
    from .meshes.mesh_truss_crane_seg3 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False
    femmesh_obj.ElementDimension = "1D"
    # four elements for each bar
    femmesh_obj.CharacteristicLengthMax = "1500.0 mm"
    femmesh_obj.CharacteristicLengthMin = "1500.0 mm"
    if FreeCAD.GuiUp:
        femmesh_obj.ViewObject.DisplayMode = "Faces, Wireframe & Nodes"
        femmesh_obj.ViewObject.PointColor = (1.0, 0.0, 0.5, 0.0)

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# -*- coding: utf8 -*-
# ***************************************************************************
# *   Copyright (c) 2014 Nathan Miller <Nathan.A.Mill@gmail.com>            *
# *   Copyright (c) 2014 Balázs Bámer                                       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************
"""The Surface Workbench GUI initialization."""

import os

import FreeCAD as App
import FreeCADGui as Gui


class SurfaceWorkbench(Gui.Workbench):
    """Surface workbench object."""

    Icon = os.path.join(
        App.getResourceDir(), "Mod", "Surface", "Resources", "icons", "Surface_Workbench.svg"
    )
    MenuText = "Surface"
    ToolTip = "Surface workbench: Create and edit complex surfaces"

    def Initialize(self):
        """Initialize the module."""
        import Surface
        import SurfaceGui

    def GetClassName(self):
        return "SurfaceGui::Workbench"


Gui.addWorkbench(SurfaceWorkbench())



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Provides a class for showing the list view and detail view at the same time."""

import base64

from addonmanager_freecad_interface import Preferences

from Addon import Addon
from Widgets.addonmanager_widget_package_details_view import PackageDetailsView
from Widgets.addonmanager_widget_view_selector import AddonManagerDisplayStyle
from addonmanager_package_details_controller import PackageDetailsController
from package_list import PackageList


from PySideWrapper import QtCore, QtWidgets


class CompositeView(QtWidgets.QWidget):
    """A widget that displays the Addon Manager's top bar, the list of Addons, and the detail
    view. Depending on the view mode selected, these may all be displayed at once, or selecting
    an addon in the list may case the list to hide and the detail view to show."""

    install = QtCore.Signal(Addon)
    uninstall = QtCore.Signal(Addon)
    update = QtCore.Signal(Addon)
    execute = QtCore.Signal(Addon)
    update_status = QtCore.Signal(Addon)
    check_for_update = QtCore.Signal(Addon)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.package_details = PackageDetailsView(self)
        self.package_details_controller = PackageDetailsController(self.package_details)
        self.package_list = PackageList(self)
        prefs = Preferences()
        self.display_style = prefs.get("ViewStyle")
        self.main_layout = QtWidgets.QHBoxLayout(self)
        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.addWidget(self.package_list)
        self.splitter.addWidget(self.package_details)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setContentsMargins(0, 0, 0, 0)
        self.splitter.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.main_layout.addWidget(self.splitter)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._setup_ui()
        self._setup_connections()
        self._restore_splitter_state()
        self.scroll_position = 0

    def _save_splitter_state(self):
        """Write the splitter state into an Addon manager preference, CompositeSplitterState"""
        prefs = Preferences()
        state = self.splitter.saveState()
        encoded = base64.b64encode(state).decode("ASCII")
        prefs.set("CompositeSplitterState", encoded)

    def _restore_splitter_state(self):
        """Restore the splitter state from CompositeSplitterState"""
        prefs = Preferences()
        encoded = prefs.get("CompositeSplitterState")
        if encoded:
            state = base64.b64decode(encoded)
            self.splitter.restoreState(state)

    def setModel(self, model):
        self.package_list.setModel(model)

    def set_display_style(self, style: AddonManagerDisplayStyle):
        self.display_style = style
        self._setup_ui()

    def _setup_ui(self):
        if self.display_style == AddonManagerDisplayStyle.EXPANDED:
            self._setup_expanded_ui()
        elif self.display_style == AddonManagerDisplayStyle.COMPACT:
            self._setup_compact_ui()
        elif self.display_style == AddonManagerDisplayStyle.COMPOSITE:
            self._setup_composite_ui()
        else:
            raise RuntimeError("Invalid display style")
        self.package_list.set_view_style(self.display_style)

    def _setup_expanded_ui(self):
        self.package_list.show()
        self.package_details.hide()
        self.package_details.button_bar.set_show_back_button(True)

    def _setup_compact_ui(self):
        self.package_list.show()
        self.package_details.hide()
        self.package_details.button_bar.set_show_back_button(True)

    def _setup_composite_ui(self):
        self.package_list.show()
        self.package_details.show()
        self.package_details.button_bar.set_show_back_button(False)

    def _setup_connections(self):
        self.package_list.itemSelected.connect(self.addon_selected)
        self.package_details_controller.back.connect(self._back_button_clicked)
        self.package_details_controller.install.connect(self.install)
        self.package_details_controller.uninstall.connect(self.uninstall)
        self.package_details_controller.update.connect(self.update)
        self.package_details_controller.execute.connect(self.execute)
        self.package_details_controller.update_status.connect(self.update_status)
        self.package_list.ui.view_bar.view_changed.connect(self.set_display_style)
        self.splitter.splitterMoved.connect(self._splitter_moved)

    def addon_selected(self, addon):
        """Depending on the display_style, show addon details (possibly hiding the package_list
        widget in the process."""
        self.package_details_controller.show_addon(addon)
        if self.display_style != AddonManagerDisplayStyle.COMPOSITE:
            self.scroll_position = (
                self.package_list.ui.listPackages.verticalScrollBar().sliderPosition()
            )
            self.package_list.hide()
            self.package_details.show()
            self.package_details.button_bar.set_show_back_button(True)

    def _back_button_clicked(self):
        if self.display_style != AddonManagerDisplayStyle.COMPOSITE:
            self.package_list.show()
            self.package_details.hide()
            # The following must be done *after* a cycle through the event loop
            QtCore.QTimer.singleShot(
                0,
                lambda: self.package_list.ui.listPackages.verticalScrollBar().setSliderPosition(
                    self.scroll_position
                ),
            )

    def _splitter_moved(self, _1: int, _2: int) -> None:
        self._save_splitter_state()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create text shapes with a particular font.

These text shapes are made of various edges and closed faces, and therefore
can be extruded to create solid bodies that can be used in boolean
operations. That is, these text shapes can be used for engraving text
into solid bodies.

They are more complex that simple text annotations.
"""
## @package gui_shapestrings
# \ingroup draftguitools
# \brief Provides GUI tools to create text shapes with a particular font.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
from draftguitools import gui_base
from draftutils import gui_utils
from draftutils.messages import _toolmsg
from draftutils.translate import translate
from drafttaskpanels import task_shapestring


class ShapeString(gui_base.GuiCommandBase):
    """Gui command for the ShapeString tool."""

    def __init__(self):
        super().__init__(name="ShapeString")

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_ShapeString",
            "MenuText": QT_TRANSLATE_NOOP("Draft_ShapeString", "Shape From Text"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_ShapeString", "Creates a shape from a text string and a specified font"
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated()
        self.ui = task_shapestring.ShapeStringTaskPanelCmd(self)
        self.call = self.view.addEventCallback("SoEvent", self.ui.action)
        _toolmsg(translate("draft", "Pick ShapeString location point"))
        task = Gui.Control.showDialog(self.ui)
        task.setDocumentName(Gui.ActiveDocument.Document.Name)
        task.setAutoCloseOnDeletedDocument(True)
        self.ui.update_hints()

    def finish(self):
        try:
            self.view.removeEventCallback("SoEvent", self.call)
            gui_utils.end_all_events()
        except RuntimeError:
            # the view has been deleted already
            pass
        self.call = None
        if Gui.Control.activeDialog():
            Gui.Control.closeDialog()
        super().finish()


Gui.addCommand("Draft_ShapeString", ShapeString())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# -*- coding: utf8 -*-
# Import gui init module
# (c) 2003 Juergen Riegel
#
# Gathering all the information to start FreeCAD
# This is the second one of three init scripts, the third one
# runs when the gui is up

# ***************************************************************************
# *   Copyright (c) 2002 Juergen Riegel <juergen.riegel@web.de>             *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/


# Registered in Part's Init.py file
FreeCAD.changeImportModule("IGES format (*.iges *.IGES *.igs *.IGS)", "Part", "ImportGui")
FreeCAD.changeExportModule("IGES format (*.iges *.igs)", "Part", "ImportGui")
FreeCAD.changeImportModule("STEP with colors (*.step *.STEP *.stp *.STP)", "Import", "ImportGui")
FreeCAD.changeExportModule("STEP with colors (*.step *.stp)", "Import", "ImportGui")
FreeCAD.changeExportModule("glTF (*.gltf *.glb)", "Import", "ImportGui")

App.__unit_test__ += ["TestImportGui"]

"""
class ImportWorkbench ( Workbench ):
    "Import workbench object"
    def Activate(self):
        # load the module
        try:
            Log ('Loading ImportGui module')
            import Import
            import ImportGui
        except ImportError:
            Err('Cannot load ImportGui')
            raise
    def GetIcon(self):
        # returns an icon for the workbench
        return ["/* XPM */\n"
            "static const char *fileopen[] = {\n"
            "\"16 13 5 1\",\n"
            "\". c #040404\",\n"
            "\"# c #808304\",\n"
            "\"a c None\",\n"
            "\"b c #f3f704\",\n"
            "\"c c #f3f7f3\",\n"
            "\"aaaaaaaaa...aaaa\",\n"
            "\"aaaaaaaa.aaa.a.a\",\n"
            "\"aaaaaaaaaaaaa..a\",\n"
            "\"a...aaaaaaaa...a\",\n"
            "\".bcb.......aaaaa\",\n"
            "\".cbcbcbcbc.aaaaa\",\n"
            "\".bcbcbcbcb.aaaaa\",\n"
            "\".cbcb...........\",\n"
            "\".bcb.#########.a\",\n"
            "\".cb.#########.aa\",\n"
            "\".b.#########.aaa\",\n"
            "\"..#########.aaaa\",\n"
            "\"...........aaaaa\"};\n"]

Gui.addWorkbench("Import",ImportWorkbench())
"""
# See https://forum.freecad.org/viewtopic.php?f=3&t=26782
# import Import_rc
# FreeCADGui.addPreferencePage(":/ui/preferences-import.ui","Import-Export")



# SPDX-License-Identifier: LGPL-2.1-or-later

#***************************************************************************
#*   Copyright (c) 2012 Sebastian Hoogen <github@sebastianhoogen.de>       *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

__title__ = "FreeCAD OpenSCAD Workbench - expand placements and matrices functions"
__author__ = "Sebastian Hoogen"
__url__ = ["https://www.freecad.org"]

'''
This Script includes python functions to shift all placements down the
feature tree to the most basic objects
'''

import FreeCAD
from OpenSCADFeatures import *
from OpenSCADUtils import isspecialorthogonal
import replaceobj


def likeprimitive(obj,extrusion=False):
    '''we can't push the matrix transformation further down'''
    return not obj.OutList or obj.isDerivedFrom('Part::Extrusion')\
        or extrusion and (obj.isDerivedFrom('Part::Revolution') \
        or obj.isDerivedFrom('Part::FeaturePython')) or \
        not obj.isDerivedFrom('Part::Feature')


def expandplacementsmatrix(obj,matrix):
    '''expand afine transformation down the feature tree'''
    ownmatrix = matrix.multiply(obj.Placement.toMatrix())
    if obj.isDerivedFrom('Part::Feature') and \
        isinstance(obj.Proxy, MatrixTransform):
        innermatrix = ownmatrix.multiply(obj.Matrix)
        if likeprimitive(obj.Base,True): #this matrix is needed
            obj.Placement = FreeCAD.Placement()
            obj.Matrix = innermatrix
        else:  #the inner object is not a primitive
            expandplacementsmatrix(obj.Base, innermatrix)
            #remove the matrix object
            for parent in obj.Base.InList:
                replaceobj.replaceobj(parent, obj, obj.Base)
            out.Document.removeObject(obj.Name)
    elif likeprimitive(obj, True):
        if isspecialorthogonal(ownmatrix):
            obj.Placement = FreeCAD.Placement()
            #this should never happen unless matrices cancel out
            obj.Placement = FreeCAD.Placement(ownmatrix)
        else:
            newobj = doc.addObject("Part::FeaturePython", 'exp_trans')
            MatrixTransform(newobj,ownmatrix,obj) #This object is not mutable GUI
            ViewProviderTree(newobj.ViewObject)
            for parent in obj.InList:
                replaceobj.replaceobj(parent, obj, newobj) # register the new object in the feature tree
            obj.Placement=FreeCAD.Placement()
    else: #not a primitive
        for outobj in obj.OutList:
            if outobj.isDerivedFrom('Part::Feature') and \
                    isinstance(obj.Proxy,MatrixTransform):
                newmatrix = ownmatrix.multiply(obj.Matrix).multiply(\
                            outobj.Base.Placement.toMatrix())
                if likeprimitive(outobj.Base,True): #child of is like primtitive
                    outobj.Matrix = newmatrix
                    outobj.Base.Placement=FreeCAD.Placement()
                else: #remove the MatrixTransformation
                    plainobj = outobj.Base
                    for parent in outobj.InList:
                        replaceobj.replaceobj(parent, outobj, plainobj)
                    outobj.Document.removeObject(outobj.Name)
                    expandplacementsmatrix(outobj,newmatrix)
            else:
                expandplacementsmatrix(outobj, ownmatrix)
        obj.Placement = FreeCAD.Placement()


def expandplacements(obj,placement):
    ownplacement = placement.multiply(obj.Placement)
    if obj.isDerivedFrom('Part::FeaturePython') and isinstance(obj.Proxy, MatrixTransform):
        expandplacementsmatrix(obj,placement.toMatrix())
    elif likeprimitive(obj, False):
        obj.Placement = ownplacement
    elif obj.isDerivedFrom('Part::Mirroring'):
        import OpenSCADUtils
        mm  = OpenSCADUtils.mirror2mat(obj.Normal,obj.Base)
        #TODO: set the base to 0,0,0
        innerp = FreeCAD.Placement(mm * ownplacement.toMatrix() *mm)
        expandplacements(obj.Source, innerp)
        obj.Placement = FreeCAD.Placement()
    else:
        for outobj in obj.OutList:
            if obj.isDerivedFrom('Part::Extrusion'):
                obj.Dir = ownplacement.Rotation.multVec(obj.Dir)
            elif obj.isDerivedFrom('Part::Revolution'):
                obj.Axis = ownplacement.Rotation.multVec(obj.Axis)
            expandplacements(outobj, ownplacement)
        obj.Placement = FreeCAD.Placement()

#expandplacements(rootobj,FreeCAD.Placement())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2021 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


"""Provides the viewprovider code for the Hatch object."""

import PySide.QtCore as QtCore
import PySide.QtGui as QtGui

import FreeCADGui as Gui

from draftguitools.gui_hatch import Draft_Hatch_TaskPanel
from draftutils.translate import translate


class ViewProviderDraftHatch:

    def __init__(self, vobj):

        vobj.Proxy = self

    def getIcon(self):

        return ":/icons/Draft_Hatch.svg"

    def dumps(self):

        return None

    def loads(self, state):

        return None

    def attach(self, vobj):

        self.Object = vobj.Object
        return

    def setEdit(self, vobj, mode):
        # EditMode 1 and 2 are handled by the Part::FeaturePython code.
        # EditMode 3 (Color) does not make sense for hatches (which do not
        # have faces) and we let that default to EditMode 0.

        if mode == 1 or mode == 2:
            return None

        taskd = Draft_Hatch_TaskPanel(vobj.Object)
        taskd.form.File.setFileName(vobj.Object.File)
        taskd.form.Pattern.setCurrentText(vobj.Object.Pattern)
        taskd.form.Scale.setValue(vobj.Object.Scale)
        taskd.form.Rotation.setValue(vobj.Object.Rotation)
        taskd.form.Translate.setChecked(vobj.Object.Translate)
        Gui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        # See setEdit.

        if mode == 1 or mode == 2:
            return None

        return True

    def setupContextMenu(self, vobj, menu):
        action_edit = QtGui.QAction(translate("draft", "Edit"), menu)
        QtCore.QObject.connect(action_edit, QtCore.SIGNAL("triggered()"), self.edit)
        menu.addAction(action_edit)

        action_transform = QtGui.QAction(
            Gui.getIcon("Std_TransformManip.svg"),
            translate("Command", "Transform"),  # Context `Command` instead of `draft`.
            menu,
        )
        QtCore.QObject.connect(action_transform, QtCore.SIGNAL("triggered()"), self.transform)
        menu.addAction(action_transform)

        return True  # Removes `Transform` and `Set colors` from the default
        # Part::FeaturePython context menu. See view_base.py.

    def edit(self):
        Gui.ActiveDocument.setEdit(self.Object, 0)

    def transform(self):
        Gui.ActiveDocument.setEdit(self.Object, 1)



# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com>   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD


# ************************************************************************************************
# setup and run examples by Python

# TODO: use method from examples gui to collect all examples in run_all method
# FreeCAD Gui update between the examples would makes sense too

"""
# setup all examples
from femexamples.manager import *
setup_all()


# run all examples
from femexamples.manager import *
run_all()


# one special example
from femexamples.manager import run_example as run

doc = run("boxanalysis_static")
doc = run("boxanalysis_frequency")

"""


def run_all():
    run_example("boxanalysis_frequency", run_solver=True)
    run_example("boxanalysis_static", run_solver=True)
    run_example("buckling_lateraltorsionalbuckling", run_solver=True)
    run_example("buckling_platebuckling", run_solver=True)
    run_example("ccx_buckling_flexuralbuckling", run_solver=True)
    run_example("ccx_cantilever_faceload", run_solver=True)
    run_example("ccx_cantilever_hexa20faceload", run_solver=True)
    run_example("ccx_cantilever_nodeload", run_solver=True)
    run_example("ccx_cantilever_prescribeddisplacement", run_solver=True)
    run_example("constraint_contact_shell_shell", run_solver=True)
    run_example("constraint_contact_solid_solid", run_solver=True)
    run_example("constraint_section_print", run_solver=True)
    run_example("constraint_selfweight_cantilever", run_solver=True)
    run_example("constraint_tie", run_solver=True)
    run_example("constraint_transform_beam_hinged", run_solver=True)
    run_example("elmer_nonguitutorial01_eigenvalue_of_elastic_beam", run_solver=True)
    run_example("equation_deformation_spring_elmer", run_solver=True)
    run_example("equation_electrostatics_capacitance_two_balls", run_solver=True)
    run_example("equation_electrostatics_electricforce_elmer_nongui6", run_solver=True)
    run_example("equation_flow_elmer_2D", run_solver=True)
    run_example("equation_flow_initial_elmer_2D", run_solver=True)
    run_example("equation_flow_turbulent_elmer_2D", run_solver=True)
    run_example("equation_flux_elmer", run_solver=True)
    run_example("equation_magnetodynamics_elmer", run_solver=True)
    run_example("equation_magnetodynamics_2D_elmer.py", run_solver=True)
    run_example("equation_magnetostatics_2D_elmer.py", run_solver=True)
    run_example("frequency_beamsimple", run_solver=True)
    run_example("material_multiple_bendingbeam_fiveboxes", run_solver=True)
    run_example("material_multiple_bendingbeam_fivefaces", run_solver=True)
    run_example("material_multiple_tensionrod_twoboxes", run_solver=True)
    run_example("material_nl_platewithhole", run_solver=True)
    run_example("rc_wall_2d", run_solver=True)
    run_example("square_pipe_end_twisted_edgeforces", run_solver=True)
    run_example("square_pipe_end_twisted_nodeforces", run_solver=True)
    run_example("thermomech_bimetal", run_solver=True)


def setup_all():
    run_example("boxanalysis_frequency")
    run_example("boxanalysis_static")
    run_example("buckling_lateraltorsionalbuckling")
    run_example("buckling_platebuckling")
    run_example("ccx_buckling_flexuralbuckling")
    run_example("ccx_cantilever_faceload")
    run_example("ccx_cantilever_hexa20faceload")
    run_example("ccx_cantilever_nodeload")
    run_example("ccx_cantilever_prescribeddisplacement")
    run_example("constraint_contact_shell_shell")
    run_example("constraint_contact_solid_solid")
    run_example("constraint_section_print")
    run_example("constraint_selfweight_cantilever")
    run_example("constraint_tie")
    run_example("constraint_transform_beam_hinged")
    run_example("elmer_nonguitutorial01_eigenvalue_of_elastic_beam")
    run_example("equation_deformation_spring_elmer")
    run_example("equation_electrostatics_capacitance_two_balls")
    run_example("equation_electrostatics_electricforce_elmer_nongui6")
    run_example("equation_flow_elmer_2D")
    run_example("equation_flow_initial_elmer_2D")
    run_example("equation_flow_turbulent_elmer_2D")
    run_example("equation_flux_elmer")
    run_example("equation_magnetodynamics_elmer")
    run_example("equation_magnetodynamics_2D_elmer.py")
    run_example("equation_magnetostatics_2D_elmer.py")
    run_example("frequency_beamsimple")
    run_example("material_multiple_bendingbeam_fiveboxes")
    run_example("material_multiple_bendingbeam_fivefaces")
    run_example("material_multiple_tensionrod_twoboxes")
    run_example("material_nl_platewithhole")
    run_example("rc_wall_2d")
    run_example("square_pipe_end_twisted_edgeforces")
    run_example("square_pipe_end_twisted_nodeforces")
    run_example("thermomech_bimetal")


def run_analysis(doc, base_name, filepath="", run_solver=False, blocking=True):

    from os.path import join, exists
    from os import makedirs
    from tempfile import gettempdir as gettmp

    # recompute
    doc.recompute()

    # print(doc.Objects)
    # print([obj.Name for obj in doc.Objects])

    # filepath
    if filepath == "":
        filepath = join(gettmp(), "FEM_examples")
    if not exists(filepath):
        makedirs(filepath)

    # find the first solver
    # thus ATM only one solver per analysis is supported
    from femtools.femutils import is_derived_from

    for m in doc.Analysis.Group:
        if is_derived_from(m, "Fem::FemSolverObjectPython"):
            solver = m
            break

    # a file name is needed for the besides dir to work
    save_fc_file = join(filepath, (base_name + ".FCStd"))
    FreeCAD.Console.PrintMessage(f"Save FreeCAD file for {base_name} analysis to {save_fc_file}\n.")
    doc.saveAs(save_fc_file)

    # get analysis workig dir
    from femtools.femutils import get_beside_dir

    working_dir = get_beside_dir(solver)

    # run analysis
    from femsolver.run import run_fem_solver

    if run_solver is True:
        run_fem_solver(solver, working_dir, blocking=blocking)

    # save doc once again with results
    doc.save()


def run_example(example, solver=None, base_name=None, run_solver=False, blocking=True):

    from importlib import import_module

    module = import_module("femexamples." + example)
    if not hasattr(module, "setup"):
        FreeCAD.Console.PrintError(f"Setup method not found in {example}\n")
        return None

    if solver is None:
        doc = getattr(module, "setup")()
    else:
        doc = getattr(module, "setup")(solvertype=solver)

    if base_name is None:
        base_name = example
        if solver is not None:
            base_name += "_" + solver
    run_analysis(doc, base_name, run_solver=run_solver, blocking=blocking)
    doc.recompute()

    return doc


# ************************************************************************************************
# helper used from examples
def init_doc(doc=None):
    if doc is None:
        doc = FreeCAD.newDocument()
        # set license
        doc.License = "Creative Commons Attribution 4.0"
        doc.LicenseURL = "https://creativecommons.org/licenses/by/4.0/"
    return doc


def get_meshname():
    # needs to be "Mesh" to work with unit tests
    return "Mesh"


def get_header(information):
    return """{name}

{information}""".format(
        name=information["name"], information=print_info_dict(information)
    )


def print_info_dict(information):
    the_text = ""
    for k, v in information.items():
        value_text = ""
        if isinstance(v, list):
            for j in v:
                value_text += f"{j}, "
            value_text = value_text.rstrip(", ")
        else:
            value_text = v
        the_text += f"{k} --> {value_text}\n"
    # print(the_text)
    return the_text


def add_explanation_obj(doc, the_text):
    text_obj = doc.addObject("App::TextDocument", "Explanation_Report")
    text_obj.Text = the_text
    text_obj.setPropertyStatus("Text", "ReadOnly")  # set property editor readonly
    if FreeCAD.GuiUp:
        text_obj.ViewObject.ReadOnly = True  # set editor view readonly
    return text_obj



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This module contains FreeCAD commands for the BIM workbench"""

import csv
import os

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")

QPROPS = [
    "Length",
    "Width",
    "Height",
    "Area",
    "HorizontalArea",
    "VerticalArea",
    "Volume",
]
TR_QPROPS = [
    translate("BIM", "Length"),
    translate("BIM", "Width"),
    translate("BIM", "Height"),
    translate("BIM", "Area"),
    translate("BIM", "Horizontal Area"),
    translate("BIM", "Vertical Area"),
    translate("BIM", "Volume"),
]
QTO_TYPES = {
    "IfcQuantityArea": "App::PropertyArea",
    "IfcQuantityCount": "App::PropertyInteger",
    "IfcQuantityLength": "App::PropertyLength",
    "IfcQuantityNumber": "App::PropertyInteger",
    "IfcQuantityTime": "App::PropertyTime",
    "IfcQuantityVolume": "App::PropertyVolume",
    "IfcQuantityWeight": "App::PropertyMass",
}


class BIM_IfcQuantities:

    def GetResources(self):
        return {
            "Pixmap": "BIM_IfcQuantities",
            "MenuText": QT_TRANSLATE_NOOP("BIM_IfcQuantities", "Manage IFC Quantities"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_IfcQuantities",
                "Manages how the quantities of different elements of the BIM project will be exported to IFC",
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        # only raise the dialog if it is already open
        if getattr(self, "form", None):
            self.form.raise_()
            return

        from PySide import QtGui

        # build objects list
        self.objectslist = {}
        self.ifcqtolist = {}
        for obj in FreeCAD.ActiveDocument.Objects:
            role = self.getRole(obj)
            if role:
                self.objectslist[obj.Name] = role
                # support for arrays
                array = self.getArray(obj)
                for i in range(array):
                    if i > 0:  # the first item already went above
                        self.objectslist[obj.Name + "+array" + str(i)] = role
        try:
            import ArchIFC

            self.ifcroles = ArchIFC.IfcTypes
        except (ImportError, AttributeError):
            import ArchComponent

            self.ifcroles = ArchComponent.IfcRoles

        # load the form and set the tree model up
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogIfcQuantities.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/BIM_IfcQuantities.svg"))
        w = PARAMS.GetInt("BimIfcQuantitiesDialogWidth", 680)
        h = PARAMS.GetInt("BimIfcQuantitiesDialogHeight", 512)
        self.form.resize(w, h)
        self.get_qtos()

        # quantities tab
        self.qmodel = QtGui.QStandardItemModel()
        self.form.quantities.setModel(self.qmodel)
        self.form.quantities.setUniformRowHeights(True)
        self.form.quantities.setItemDelegate(QtGui.QStyledItemDelegate())
        self.qmodel.dataChanged.connect(self.setChecked)
        self.form.buttonBox.accepted.connect(self.accept)
        self.form.rejected.connect(self.reject)  # also triggered by self.form.buttonBox.rejected
        self.form.quantities.clicked.connect(self.onClickTree)
        if hasattr(self.form.onlyVisible, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.onlyVisible.checkStateChanged.connect(self.update)
        else:  # Qt version < 6.7.0
            self.form.onlyVisible.stateChanged.connect(self.update)
        self.form.buttonRefresh.clicked.connect(self.update)
        self.form.buttonApply.clicked.connect(self.add_qto)

        # center the dialog over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.form.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.form.rect().center()
        )

        self.update()
        self.form.show()

    def getArray(self, obj):
        "returns a count number if this object needs to be duplicated"

        import Draft

        if len(obj.InList) == 1:
            parent = obj.InList[0]
            if Draft.getType(parent) == "Array":
                return parent.Count
        return 0

    def decamelize(self, s):
        return "".join([" " + c if c.isupper() else c for c in s]).strip(" ")

    def get_qtos(self):
        "populates the qtos combo box"

        def read_csv(csvfile):
            result = {}
            if os.path.exists(csvfile):
                with open(csvfile, "r") as f:
                    reader = csv.reader(f, delimiter=";")
                    for row in reader:
                        result[row[0]] = row[1:]
            return result

        self.qtodefs = {}
        qtopath = os.path.join(
            FreeCAD.getResourceDir(), "Mod", "BIM", "Presets", "qto_definitions.csv"
        )
        custompath = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "CustomQtos.csv")
        self.qtodefs = read_csv(qtopath)
        self.qtodefs.update(read_csv(custompath))
        self.qtokeys = [
            "".join(map(lambda x: x if x.islower() else " " + x, t[4:]))[1:]
            for t in self.qtodefs.keys()
        ]
        self.qtokeys.sort()
        self.form.comboQto.addItems(
            [
                translate("BIM", "Add quantity set..."),
            ]
            + self.qtokeys
        )

    def add_qto(self):
        "Adds a standard qto set to the todo list"

        index = self.form.comboQto.currentIndex()
        if index <= 0:
            return
        if len(FreeCADGui.Selection.getSelection()) != 1:
            return
        obj = FreeCADGui.Selection.getSelection()[0]
        qto = list(self.qtodefs.keys())[index - 1]
        self.ifcqtolist.setdefault(obj.Name, []).append(qto)
        self.update_line(obj.Name, qto)
        FreeCAD.Console.PrintMessage(translate("BIM", "Adding quantity set") + ": " + qto + "\n")

    def apply_qto(self, obj, qto):
        "Adds a standard qto set to the object"

        val = self.qtodefs[qto]
        qset = None
        if hasattr(obj, "StepId"):
            from nativeifc import ifc_tools

            ifcfile = ifc_tools.get_ifcfile(obj)
            element = ifc_tools.get_ifc_element(obj)
            if not ifcfile or not element:
                return
            qset = ifc_tools.api_run("pset.add_qto", ifcfile, product=element, name=qto)
        for i in range(0, len(val), 2):
            qname = val[i]
            qtype = QTO_TYPES[val[i + 1]]
            if not qname in obj.PropertiesList:
                obj.addProperty(qtype, qname, "Quantities", val[i + 1], locked=True)
                qval = 0
                i = self.get_row(obj.Name)
                if i > -1:
                    for j, p in enumerate(QPROPS):
                        it = self.qmodel.item(i, j + 1)
                        t = it.text()
                        if t:
                            t = t.replace("²", "^2").replace("³", "^3")
                            qval = FreeCAD.Units.Quantity(t).Value
                if qval:
                    setattr(obj, qname, qval)
                if hasattr(obj, "StepId") and qset:
                    ifc_tools.api_run("pset.edit_qto", ifcfile, qto=qset, properties={qname: qval})

    def update(self, index=None):
        """updates the tree widgets in all tabs. Index is not used,
        it is just there to match a qt slot requirement"""

        from PySide import QtCore, QtGui
        import Draft

        # quantities tab

        self.qmodel.clear()
        self.qmodel.setHorizontalHeaderLabels([translate("BIM", "Label")] + TR_QPROPS)
        self.form.quantities.setColumnWidth(0, 200)  # TODO remember width
        quantheaders = self.form.quantities.header()  # QHeaderView instance
        quantheaders.setSectionsClickable(True)
        quantheaders.sectionClicked.connect(self.quantHeaderClicked)

        # sort by type

        groups = {}
        for name, role in self.objectslist.items():
            groups.setdefault(role, []).append(name)
        for names in groups.values():
            suffix = ""
            for name in names:
                if "+array" in name:
                    name = name.split("+array")[0]
                    suffix = " (duplicate)"
                obj = FreeCAD.ActiveDocument.getObject(name)
                if obj:
                    if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                        if obj.isDerivedFrom("Part::Feature") and not (
                            Draft.getType(obj) == "Site"
                        ):
                            it1 = QtGui.QStandardItem(obj.Label + suffix)
                            it1.setToolTip(name + suffix)
                            it1.setEditable(False)
                            it1.setIcon(obj.ViewObject.Icon)
                            props = []
                            for prop in QPROPS:
                                it = QtGui.QStandardItem()
                                val = None
                                if hasattr(obj, prop) and ("Hidden" not in obj.getEditorMode(prop)):
                                    val = self.get_text(obj, prop)
                                    it.setText(val)
                                    it.setCheckable(True)
                                if val != None:
                                    d = None
                                    if hasattr(obj, "IfcAttributes"):
                                        d = obj.IfcAttributes
                                    elif hasattr(obj, "IfcData"):
                                        d = obj.IfcData
                                    if d:
                                        if ("Export" + prop in d) and (
                                            d["Export" + prop] == "True"
                                        ):
                                            it.setCheckState(QtCore.Qt.Checked)
                                    elif self.has_qto(obj, prop):
                                        it.setCheckState(QtCore.Qt.Checked)
                                    if val == 0:
                                        it.setIcon(QtGui.QIcon(":/icons/warning.svg"))
                                self.set_editable(it, prop)
                                props.append(it)
                            self.qmodel.appendRow([it1] + props)

    def has_qto(self, obj, prop):
        """Says if the given object has the given prop in a qto set"""

        if not "StepId" in obj.PropertiesList:
            return False
        from nativeifc import ifc_tools

        element = ifc_tools.get_ifc_element(obj)
        if not element:
            return False
        for rel in getattr(element, "IsDefinedBy", []):
            pset = rel.RelatingPropertyDefinition
            if pset.is_a("IfcElementQuantity"):
                if pset.Name in self.qtodefs:
                    if prop in self.qtodefs[pset.Name]:
                        return True
        return False

    def get_text(self, obj, prop):
        """Gets the text from a property"""

        val = getattr(obj, prop, "0")
        txt = val.getUserPreferred()[0].replace("^2", "²").replace("^3", "³")
        return txt

    def get_row(self, name):
        """Returns the row number corresponding to the given object name"""

        for i in range(self.qmodel.rowCount()):
            if self.qmodel.item(i).toolTip().split(" ")[0] == name:
                return i
        return -1

    def update_line(self, name, qto):
        """Updates a single line of the table, without updating
        the actual object"""

        from PySide import QtCore

        i = self.get_row(name)
        if i == -1:
            return
        obj = FreeCAD.ActiveDocument.getObject(name)
        qto_val = self.qtodefs[qto]
        for j, p in enumerate(QPROPS):
            it = self.qmodel.item(i, j + 1)
            if p in obj.PropertiesList:
                val = self.get_text(obj, p)
                it.setText(val)
                self.set_editable(it, p)
                it.setCheckable(True)
            elif p in qto_val:
                it.setText("0")
                it.setCheckable(True)
                it.setCheckState(QtCore.Qt.Checked)
                self.set_editable(it, p)

    def set_editable(self, it, prop):
        """Checks if the given prop should be editable, and sets it"""

        if prop in ["Area", "HorizontalArea", "VerticalArea", "Volume"]:
            it.setEditable(False)
        else:
            it.setEditable(True)

    def getRole(self, obj):
        """gets the IFC class of this object"""

        if hasattr(obj, "IfcType"):
            return obj.IfcType
        elif hasattr(obj, "IfcRole"):
            return obj.IfcRole
        elif hasattr(obj, "IfcClass"):
            return obj.IfcClass
        else:
            return None

    def accept(self):
        """OK pressed"""

        PARAMS.SetInt("BimIfcQuantitiesDialogWidth", self.form.width())
        PARAMS.SetInt("BimIfcQuantitiesDialogHeight", self.form.height())
        self.form.hide()
        changed = False
        if self.ifcqtolist:
            if not changed:
                FreeCAD.ActiveDocument.openTransaction("Change quantities")
            changed = True
            for key, val in self.ifcqtolist.items():
                obj = FreeCAD.ActiveDocument.getObject(key)
                if obj:
                    for qto in val:
                        self.apply_qto(obj, qto)
        for row in range(self.qmodel.rowCount()):
            name = self.qmodel.item(row, 0).toolTip()
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                for i in range(len(QPROPS)):
                    item = self.qmodel.item(row, i + 1)
                    val = item.text()
                    sav = bool(item.checkState())
                    if i < 3:  # Length, Width, Height, value can be changed
                        if hasattr(obj, QPROPS[i]):
                            if getattr(obj, QPROPS[i]).getUserPreferred()[0] != val:
                                setattr(obj, QPROPS[i], val)
                                if not changed:
                                    FreeCAD.ActiveDocument.openTransaction("Change quantities")
                                changed = True
                    d = None
                    if hasattr(obj, "IfcAttributes"):
                        d = obj.IfcAttributes
                        att = "IfcAttributes"
                    elif hasattr(obj, "IfcData"):
                        d = obj.IfcData
                        att = "IfcData"
                    if d:
                        if sav:
                            if (not "Export" + QPROPS[i] in d) or (
                                d["Export" + QPROPS[i]] == "False"
                            ):
                                d["Export" + QPROPS[i]] = "True"
                                setattr(obj, att, d)
                                if not changed:
                                    FreeCAD.ActiveDocument.openTransaction("Change quantities")
                                changed = True
                        else:
                            if "Export" + QPROPS[i] in d:
                                if d["Export" + QPROPS[i]] == "True":
                                    d["Export" + QPROPS[i]] = "False"
                                    setattr(obj, att, d)
                                    if not changed:
                                        FreeCAD.ActiveDocument.openTransaction("Change quantities")
                                    changed = True
                    elif "StepId" not in obj.PropertiesList:
                        FreeCAD.Console.PrintError(
                            translate(
                                "BIM", "Cannot save quantities settings for object %1"
                            ).replace("%1", obj.Label)
                            + "\n"
                        )

        if changed:
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()

        return self.reject()

    def reject(self):
        self.form.hide()
        del self.form
        return True

    def setChecked(self, id1, id2):
        sel = self.form.quantities.selectedIndexes()
        state = self.qmodel.itemFromIndex(id1).checkState()
        if len(sel) > 7:
            for idx in sel:
                if idx.column() == id1.column():
                    item = self.qmodel.itemFromIndex(idx)
                    if item.checkState() != state:
                        item.setCheckState(state)

    def quantHeaderClicked(self, col):
        from PySide import QtCore

        sel = self.form.quantities.selectedIndexes()
        state = None
        if len(sel) > 7:
            for idx in sel:
                if idx.column() == col:
                    item = self.qmodel.itemFromIndex(idx)
                    if state is None:
                        # take the state to apply from the first selected item
                        state = QtCore.Qt.Checked
                        if item.checkState():
                            state = QtCore.Qt.Unchecked
                    item.setCheckState(state)

    def onClickTree(self, index=None):

        FreeCADGui.Selection.clearSelection()
        sel = self.form.quantities.selectedIndexes()
        for index in sel:
            if index.column() == 0:
                obj = FreeCAD.ActiveDocument.getObject(self.qmodel.itemFromIndex(index).toolTip())
                if obj:
                    FreeCADGui.Selection.addSelection(obj)


FreeCADGui.addCommand("BIM_IfcQuantities", BIM_IfcQuantities())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide import QtGui
import FreeCADGui
import Path
from Path.Base.Gui import IconViewProvider
from Path.Tool.toolbit.ui.panel import TaskPanel


class ViewProvider(object):
    """
    ViewProvider for a ToolBit DocumentObject.
    It's sole job is to provide an icon and invoke the TaskPanel
    on edit.
    """

    def __init__(self, vobj, name):
        Path.Log.track(name, vobj.Object)
        self.panel = None
        self.icon = name
        self.obj = vobj.Object
        self.vobj = vobj
        vobj.Proxy = self

    def attach(self, vobj):
        Path.Log.track(vobj.Object)
        self.vobj = vobj
        self.obj = vobj.Object

    def getIcon(self):
        try:
            png_data = self.obj.Proxy.get_thumbnail()
        except AttributeError:  # Proxy not initialized
            png_data = None
        if png_data:
            pixmap = QtGui.QPixmap()
            pixmap.loadFromData(png_data, "PNG")
            return QtGui.QIcon(pixmap)
        return ":/icons/CAM_ToolBit.svg"

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onDelete(self, vobj, arg2=None):
        Path.Log.track(vobj.Object.Label)
        vobj.Object.Proxy.onDelete(vobj.Object)

    def getDisplayMode(self, mode):
        return "Default"

    def _openTaskPanel(self, vobj, deleteOnReject):
        Path.Log.track()
        self.panel = TaskPanel(vobj, deleteOnReject)
        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(self.panel)
        self.panel.setupUi()

    def setCreate(self, vobj):
        Path.Log.track()
        self._openTaskPanel(vobj, True)

    def setEdit(self, vobj, mode=0):
        self._openTaskPanel(vobj, False)
        return True

    def unsetEdit(self, vobj, mode):
        FreeCADGui.Control.closeDialog()
        self.panel = None
        return

    def claimChildren(self):
        if self.obj.BitBody:
            return [self.obj.BitBody]
        return []

    def doubleClicked(self, vobj):
        pass

    def setupContextMenu(self, vobj, menu):
        # Override the base class method to prevent adding the "Edit" action
        # for ToolBit objects.
        pass  # TODO: call setEdit here once we have a new editor panel


IconViewProvider.RegisterViewProvider("ToolBit", ViewProvider)
